#!/usr/env/python

"""
fill_sinks_barnes.py

Fill sinks in a landscape to the brim, following the Barnes et al. (2014)
algorithms.
"""

from __future__ import print_function

import numpy as np
from six import iteritems

from landlab import BAD_INDEX_VALUE
from landlab.components import LakeMapperBarnes
from landlab.utils.return_array import return_array_at_node

LOCAL_BAD_INDEX_VALUE = BAD_INDEX_VALUE


class SinkFillerBarnes(LakeMapperBarnes):
    """
    Uses the Barnes et al (2014) algorithms to replace pits in a topography
    with flats, or optionally with very shallow gradient surfaces to allow
    continued draining.

    This component is NOT intended for use iteratively as a model runs;
    rather, it is to fill in an initial topography. If you want to repeatedly
    fill pits as a landscape develops, you are after the LakeMapperBarnes
    component. If you want flow paths on your filled landscape, manually run a
    FlowDirector and FlowAccumulator for yourself.

    The locations and depths etc. of the fills will be tracked, and properties
    are provided to access this information.

    Parameters
    ----------
    grid : ModelGrid
        A grid.
    surface : field name at node or array of length node
        The surface to fill.
    method : {'Steepest', 'D8'}
        Whether or not to recognise diagonals as valid flow paths, if a raster.
        Otherwise, no effect.
    fill_flat : bool
        If True, pits will be filled to perfectly horizontal. If False, the new
        surface will be slightly inclined (at machine precision) to give
        steepest descent flow paths to the outlet, once they are calculated.
    ignore_overfill : bool
        If True, suppresses the Error that would normally be raised during
        creation of a gentle incline on a fill surface (i.e., if not
        fill_flat). Typically this would happen on a synthetic DEM where more
        than one outlet is possible at the same elevation. If True, the
        was_there_overfill property can still be used to see if this has
        occurred.
    """

    _name = "SinkFillerBarnes"

    _input_var_names = ("topographic__elevation",)

    _output_var_names = ("topographic__elevation", "sediment_fill__depth")

    _var_units = {"topographic__elevation": "m", "sediment_fill__depth": "m"}

    _var_mapping = {"topographic__elevation": "node", "sediment_fill__depth": "node"}

    _var_doc = {
        "topographic__elevation": "Surface topographic elevation",
        "sediment_fill__depth": "Depth of sediment added at each" + "node",
    }

    def __init__(
        self,
        grid,
        surface="topographic__elevation",
        method="D8",
        fill_flat=False,
        ignore_overfill=False,
    ):
        """
        Initialise the component.
        """
        if "flow__receiver_node" in grid.at_node:
            if grid.at_node["flow__receiver_node"].size != grid.size("node"):
                msg = (
                    "A route-to-multiple flow director has been "
                    "run on this grid. The landlab development team has not "
                    "verified that SinkFillerBarnes is compatible with "
                    "route-to-multiple methods. Please open a GitHub Issue "
                    "to start this process."
                )
                raise NotImplementedError(msg)

        # Most of the functionality of this component is directly inherited
        # from SinkFillerBarnes, so
        super(SinkFillerBarnes, self).__init__(
            grid,
            surface=surface,
            method=method,
            fill_flat=fill_flat,
            fill_surface=surface,
            redirect_flow_steepest_descent=False,
            reaccumulate_flow=False,
            ignore_overfill=ignore_overfill,
            track_lakes=True,
        )
        # note we will always track the fills, since we're only doing this
        # once... Likewise, no need for flow routing; this is not going to
        # get used dynamically.
        self._supplied_surface = return_array_at_node(grid, surface).copy()
        # create the only new output field:
        self._sed_fill_depth = self.grid.add_zeros(
            "node", "sediment_fill__depth", noclobber=False
        )

    def run_one_step(self):
        """
        Fills the surface to remove all pits.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab import RasterModelGrid, CLOSED_BOUNDARY
        >>> from landlab.components import SinkFillerBarnes, FlowAccumulator
        >>> mg = RasterModelGrid((5, 6))
        >>> for edge in ('left', 'top', 'bottom'):
        ...     mg.status_at_node[mg.nodes_at_edge(edge)] = CLOSED_BOUNDARY
        >>> z = mg.add_zeros('node', 'topographic__elevation', dtype=float)
        >>> z.reshape(mg.shape)[2, 1:-1] = [2., 1., 0.5, 1.5]
        >>> z.reshape(mg.shape)[1, 1:-1] = [2.1, 1.1, 0.6, 1.6]
        >>> z.reshape(mg.shape)[3, 1:-1] = [2.2, 1.2, 0.7, 1.7]
        >>> z_init = z.copy()
        >>> sfb = SinkFillerBarnes(mg, method='Steepest')  #, surface=z

        TODO: once return_array_at_node is fixed, this example should also
        take surface... GIVE IT surface=z  !!

        >>> sfb.run_one_step()
        >>> z_out = np.array([ 0. ,  0. ,  0. ,  0. ,  0. ,  0. ,
        ...                    0. ,  2.1,  1.5,  1.5,  1.6,  0. ,
        ...                    0. ,  2. ,  1.5,  1.5,  1.5,  0. ,
        ...                    0. ,  2.2,  1.5,  1.5,  1.7,  0. ,
        ...                    0. ,  0. ,  0. ,  0. ,  0. ,  0. ])
        >>> # np.allclose(z, z_out)  ->  True once fixed
        >>> np.all(np.equal(z, z_out))  # those 1.5's are actually a bit > 1.5
        False
        >>> fill_map = np.array([-1, -1, -1, -1, -1, -1,
        ...                      -1, -1, 16, 16, -1, -1,
        ...                      -1, -1, 16, 16, -1, -1,
        ...                      -1, -1, 16, 16, -1, -1,
        ...                      -1, -1, -1, -1, -1, -1])
        >>> np.all(sfb.fill_map == fill_map)
        True
        >>> np.all(sfb.fill_at_node == (sfb.fill_map > -1))  # bool equivalent
        True
        >>> sfb.was_there_overfill  # everything fine with slope adding
        False

        >>> fr = FlowAccumulator(mg, flow_director='D4')  # routing now works
        >>> fr.run_one_step()
        >>> np.all(mg.at_node['flow__sink_flag'][mg.core_nodes] == 0)
        True
        >>> drainage_area = np.array([  0.,   0.,   0.,   0.,   0.,   0.,
        ...                             0.,   1.,   2.,   3.,   1.,   1.,
        ...                             0.,   1.,   4.,   9.,  10.,  10.,
        ...                             0.,   1.,   2.,   1.,   1.,   1.,
        ...                             0.,   0.,   0.,   0.,   0.,   0.])
        >>> np.allclose(mg.at_node['drainage_area'], drainage_area)
        True

        Test two pits and a flat fill:

        >>> from collections import deque
        >>> z[:] = mg.node_x.max() - mg.node_x
        >>> z[[10, 23]] = 1.1  # raise "guard" exit nodes
        >>> z[7] = 2.  # is a lake on its own
        >>> z[9] = 0.5
        >>> z[15] = 0.3
        >>> z[14] = 0.6  # [9, 14, 15] is a lake
        >>> z[22] = 0.9  # a non-contiguous lake node also draining to 16
        >>> z_init = z.copy()
        >>> sfb = SinkFillerBarnes(mg, method='Steepest', fill_flat=True)
        >>> sfb.run_one_step()
        >>> sfb.fill_dict  == {8: deque([7]), 16: deque([15, 9, 14, 22])}
        True
        >>> sfb.number_of_fills
        2
        >>> sfb.fill_outlets == [16, 8]
        True
        >>> np.allclose(sfb.fill_areas, np.array([4., 1.]))  # same order
        True

        Unlike the LakeMapperBarnes equivalents, fill_depths and fill_volumes
        are always available through this component:

        >>> fill_depths = np.array([ 0. ,  0. ,  0. ,  0. ,  0. ,  0. ,
        ...                          0. ,  1. ,  0. ,  0.5,  0. ,  0. ,
        ...                          0. ,  0. ,  0.4,  0.7,  0. ,  0. ,
        ...                          0. ,  0. ,  0. ,  0. ,  0.1,  0. ,
        ...                          0. ,  0. ,  0. ,  0. ,  0. ,  0. ])
        >>> np.allclose(sfb.fill_depths, fill_depths)
        True
        >>> np.allclose(sfb.fill_volumes, np.array([1.7, 1.]))
        True

        Note that with a flat fill, we can't drain the surface. The surface
        is completely flat, so each and every core node within the fill
        becomes a sink.

        >>> fr.run_one_step()
        >>> where_is_filled = np.where(sfb.fill_map > -1, 1, 0)
        >>> np.all(
        ...     mg.at_node['flow__sink_flag'][mg.core_nodes] ==
        ...     where_is_filled[mg.core_nodes])
        True

        (Note that the fill_map does not think that the perimeter nodes are
        sinks, since they haven't changed elevation. In contrast, the
        FlowAccumulator *does* think they are, because these nodes are where
        flow terminates.)
        """
        if "flow__receiver_node" in self._grid.at_node:
            if self._grid.at_node["flow__receiver_node"].size != self._grid.size(
                "node"
            ):
                msg = (
                    "A route-to-multiple flow director has been "
                    "run on this grid. The landlab development team has "
                    "not verified that SinkFillerBarnes is compatible with "
                    "route-to-multiple methods. Please open a GitHub Issue "
                    "to start this process."
                )
                raise NotImplementedError(msg)
        super(SinkFillerBarnes, self).run_one_step()
        self._sed_fill_depth[:] = self._surface - self._supplied_surface

    @property
    def fill_dict(self):
        """
        Return a dictionary where the keys are the outlet nodes of each filled
        area, and the values are deques of nodes within each. Items are not
        returned in ID order.
        """
        return super(SinkFillerBarnes, self).lake_dict

    @property
    def fill_outlets(self):
        """
        Returns the outlet for each filled area, not necessarily in ID order.
        """
        return super(SinkFillerBarnes, self).lake_outlets

    @property
    def number_of_fills(self):
        """
        Return the number of individual filled areas.
        """
        return super(SinkFillerBarnes, self).number_of_lakes

    @property
    def fill_map(self):
        """
        Return an array of ints, where each filled node is labelled
        with its outlet node ID.
        Nodes not in a filled area are labelled with LOCAL_BAD_INDEX_VALUE
        (default -1).
        """
        return super(SinkFillerBarnes, self).lake_map

    @property
    def fill_at_node(self):
        """
        Return a boolean array, True if the node is filled, False otherwise.
        """
        return super(SinkFillerBarnes, self).lake_at_node

    @property
    def fill_depths(self):
        """Return the change in surface elevation at each node this step.
        """
        return self._sed_fill_depth

    @property
    def fill_areas(self):
        """
        A nlakes-long array of the area of each fill. The order is the same as
        that of the keys in fill_dict, and of fill_outlets.
        """
        return super(SinkFillerBarnes, self).lake_areas

    @property
    def fill_volumes(self):
        """
        A nlakes-long array of the volume of each fill. The order is the same
        as that of the keys in fill_dict, and of fill_outlets.
        """
        fill_vols = np.empty(self.number_of_fills, dtype=float)
        col_vols = self.grid.cell_area_at_node * self.fill_depths
        for (i, (outlet, fillnodes)) in enumerate(iteritems(self.fill_dict)):
            fill_vols[i] = col_vols[fillnodes].sum()
        return fill_vols



#!/usr/bin/env python
import numpy as np
import pytest

from landlab import RasterModelGrid
from landlab.components import FlowAccumulator
from landlab.utils import (
    get_watershed_mask,
    get_watershed_masks_with_area_threshold,
    get_watershed_nodes,
    get_watershed_outlet,
)


def test_get_watershed_nodes():
    grid = RasterModelGrid((7, 7))

    z = np.array(
        [
            -9999.0,
            -9999.0,
            -9999.0,
            -9999.0,
            -9999.0,
            -9999.0,
            -9999.0,
            -9999.0,
            26.0,
            0.0,
            30.0,
            32.0,
            34.0,
            -9999.0,
            -9999.0,
            28.0,
            1.0,
            25.0,
            28.0,
            32.0,
            -9999.0,
            -9999.0,
            30.0,
            3.0,
            3.0,
            11.0,
            34.0,
            -9999.0,
            -9999.0,
            32.0,
            11.0,
            25.0,
            18.0,
            38.0,
            -9999.0,
            -9999.0,
            34.0,
            32.0,
            34.0,
            36.0,
            40.0,
            -9999.0,
            -9999.0,
            -9999.0,
            -9999.0,
            -9999.0,
            -9999.0,
            -9999.0,
            -9999.0,
        ]
    )

    grid.at_node["topographic__elevation"] = z

    outlet_id = 2

    grid.set_watershed_boundary_condition_outlet_id(outlet_id, z, nodata_value=-9999.0)

    fr = FlowAccumulator(grid, flow_director="D8")
    fr.run_one_step()

    ws_nodes = get_watershed_nodes(grid, outlet_id)

    # Given the watershed boundary conditions, the number of watershed nodes
    # should be equal to the number of core nodes plus 1 for the outlet node.
    np.testing.assert_equal(len(ws_nodes), grid.number_of_core_nodes + 1)


def test_get_watershed_masks_with_area_threshold():
    rmg = RasterModelGrid((7, 7), xy_spacing=200)

    z = np.array(
        [
            -9999.0,
            -9999.0,
            -9999.0,
            -9999.0,
            -9999.0,
            -9999.0,
            -9999.0,
            -9999.0,
            26.0,
            0.0,
            26.0,
            30.0,
            34.0,
            -9999.0,
            -9999.0,
            28.0,
            1.0,
            28.0,
            5.0,
            32.0,
            -9999.0,
            -9999.0,
            30.0,
            3.0,
            30.0,
            10.0,
            34.0,
            -9999.0,
            -9999.0,
            32.0,
            11.0,
            32.0,
            15.0,
            38.0,
            -9999.0,
            -9999.0,
            34.0,
            32.0,
            34.0,
            36.0,
            40.0,
            -9999.0,
            -9999.0,
            -9999.0,
            -9999.0,
            -9999.0,
            -9999.0,
            -9999.0,
            -9999.0,
        ]
    )

    rmg.at_node["topographic__elevation"] = z

    rmg.set_closed_boundaries_at_grid_edges(True, True, True, False)

    # Route flow.
    fr = FlowAccumulator(rmg, flow_director="D8")
    fr.run_one_step()

    # Get the masks of watersheds greater than or equal to 80,000
    # square-meters.
    critical_area = 80000
    mask = get_watershed_masks_with_area_threshold(rmg, critical_area)

    # Assert that mask null nodes have a drainage area below critical area.
    null_nodes = np.where(mask == -1)[0]
    A = rmg.at_node["drainage_area"][null_nodes]
    below_critical_area_nodes = A < critical_area
    trues = np.ones(len(A), dtype=bool)

    np.testing.assert_equal(below_critical_area_nodes, trues)


def test_get_watershed_outlet():
    grid = RasterModelGrid((7, 7))

    z = np.array(
        [
            -9999.0,
            -9999.0,
            -9999.0,
            -9999.0,
            -9999.0,
            -9999.0,
            -9999.0,
            -9999.0,
            26.0,
            0.0,
            30.0,
            32.0,
            34.0,
            -9999.0,
            -9999.0,
            28.0,
            1.0,
            25.0,
            28.0,
            32.0,
            -9999.0,
            -9999.0,
            30.0,
            3.0,
            3.0,
            11.0,
            34.0,
            -9999.0,
            -9999.0,
            32.0,
            11.0,
            25.0,
            18.0,
            38.0,
            -9999.0,
            -9999.0,
            34.0,
            32.0,
            34.0,
            36.0,
            40.0,
            -9999.0,
            -9999.0,
            -9999.0,
            -9999.0,
            -9999.0,
            -9999.0,
            -9999.0,
            -9999.0,
        ]
    )

    grid.at_node["topographic__elevation"] = z

    imposed_outlet = 2

    grid.set_watershed_boundary_condition_outlet_id(
        imposed_outlet, z, nodata_value=-9999.0
    )

    fr = FlowAccumulator(grid, flow_director="D8")
    fr.run_one_step()

    test_node = 32

    determined_outlet = get_watershed_outlet(grid, test_node)
    np.testing.assert_equal(determined_outlet, imposed_outlet)

    # Create a pit.
    pit_node = 38
    grid.at_node["topographic__elevation"][pit_node] -= 32
    fr.run_one_step()

    pit_outlet = get_watershed_outlet(grid, test_node)
    np.testing.assert_equal(pit_outlet, pit_node)


def test_route_to_multiple_error_raised_watershed_outlet():
    mg = RasterModelGrid((10, 10))
    z = mg.add_zeros("node", "topographic__elevation")
    z += mg.x_of_node + mg.y_of_node
    fa = FlowAccumulator(mg, flow_director="MFD")
    fa.run_one_step()

    with pytest.raises(NotImplementedError):
        get_watershed_outlet(mg, 10)


def test_route_to_multiple_error_raised_watershed_mask():
    mg = RasterModelGrid((10, 10))
    z = mg.add_zeros("node", "topographic__elevation")
    z += mg.x_of_node + mg.y_of_node
    fa = FlowAccumulator(mg, flow_director="MFD")
    fa.run_one_step()

    with pytest.raises(NotImplementedError):
        get_watershed_mask(mg, 10)



"""
========================================================================================
Wrap landlab component with the Basic Modeling Interface (:mod:`landlab.bmi.bmi_bridge`)
========================================================================================

.. sectionauthor:: Eric Hutton

Function reference
------------------

The `wrap_as_bmi` function wraps a landlab component class so that it
exposes a Basic Modelling Interface.

"""
import inspect

import numpy as np
from bmipy import Bmi

from ..core import load_params
from ..core.model_component import Component
from ..framework.decorators import snake_case
from ..grid import HexModelGrid, RasterModelGrid
from ..grid.create import create_grid

BMI_LOCATION = {
    "node": "node",
    "link": "edge",
    "patch": "face",
    "corner": "node",
    "face": "edge",
    "cell": "face",
    "grid": "none",
}

BMI_GRID = {
    "node": 0,
    "link": 0,
    "patch": 0,
    "corner": 1,
    "face": 1,
    "cell": 1,
    "grid": 2,
}


class TimeStepper(object):

    """Step through time.

    Parameters
    ----------
    start : float, optional
        Clock start time.
    stop : float, optional
        Stop time.
    step : float, optional
        Time step.

    Examples
    --------
    >>> from landlab.bmi import TimeStepper
    >>> time_stepper = TimeStepper()
    >>> time_stepper.start
    0.0
    >>> time_stepper.stop is None
    True
    >>> time_stepper.step
    1.0
    >>> time_stepper.time
    0.0
    >>> for _ in range(10): time_stepper.advance()
    >>> time_stepper.time
    10.0
    >>> time_stepper = TimeStepper(1., 13., 2.)
    >>> [time for time in time_stepper]
    [1.0, 3.0, 5.0, 7.0, 9.0, 11.0]
    """

    def __init__(self, start=0.0, stop=None, step=1.0, units="s"):
        self._start = start
        self._stop = stop
        self._step = step
        self._units = units

        self._time = start

    def __iter__(self):
        if self.stop is None:
            while 1:
                yield self._time
                self._time += self._step
        else:
            while self._time < self._stop:
                yield self._time
                self._time += self._step
        return

    @property
    def time(self):
        """Current time."""
        return self._time

    @property
    def start(self):
        """Start time."""
        return self._start

    @property
    def stop(self):
        """Stop time."""
        return self._stop

    @property
    def step(self):
        """Time Step."""
        return self._step

    @step.setter
    def step(self, new_val):
        """Change the time step."""
        self._step = new_val

    @property
    def units(self):
        """Time units."""
        return self._units

    def advance(self):
        """Advance the time stepper by one time step."""
        self._time += self.step
        if self._stop is not None and self._time > self._stop:
            raise StopIteration()


def wrap_as_bmi(cls):
    """Wrap a landlab class so it exposes a BMI.

    Give a landlab component a Basic Model Interface (BMI). Since landlab
    components have an interface that is already in the style of BMI,
    this function adds just a light wrapping to landlab components. There
    are a number of differences that may cause some confusion to
    landlab users.

    1.  Because BMI doesn't have a concept of a dual grid, it only
        defines *nodes* (points), *edges* (vectors), and *faces*
        (areas). The dual-graph of landlab is considered as two
        separate grids by BMI.

    2. It is important to note that BMI has only three grid elements
       (*node*, *edge*, and *face*) while landlab has 6. The names
       used by landlab and BMI are also different.

       Thus, a BMI-wrapped landlab component will always have two
       grids with grid identifiers 0, and 1. Grid 0 will contain
       the landlab *nodes*, *links*, and *patches* while grid 1 will
       contain *corners*, *faces*, and *cells*.landlab and BMI
       refer to grid elements by different names. The mapping from
       landlab to BMI nomenclature is the following:

        Grid 0:
        *  *node*: *node*
        *  *link*: *edge*
        *  *patch*: *face*

        Grid 1:
        *  *corner*: *node*
        *  *face*: *edge*
        *  *cell*: *face*

    3.  In BMI, the *initialize* method requires an input file that is
        used to create and setup the model for time-stepping. landlab
        components generally do not have anything like this; instead
        this task is usually done programmatically. Thus, the
        input file that is used by the BMI *initialize* method is
        a standard landlab input file as used by the landlab *create_grid*
        function.

    Parameters
    ----------
    cls : class
        A landlab class that inherits from `Component`.

    Returns
    -------
    class
        A wrapped class that exposes a BMI.

    Examples
    --------
    >>> from landlab.bmi import wrap_as_bmi
    >>> from landlab.components.flexure import Flexure

    >>> BmiFlexure = wrap_as_bmi(Flexure)
    >>> flexure = BmiFlexure()

    >>> config = \"\"\"
    ... flexure:
    ...     eet: 10.e+3
    ...     method: flexure
    ... clock:
    ...     start: 0.
    ...     stop: 10.
    ...     step: 2.
    ... grid:
    ...     RasterModelGrid:
    ...     - [20, 40]
    ...     - xy_spacing: [2000., 1000.]
    ... \"\"\"
    >>> flexure.initialize(config)
    >>> sorted(flexure.get_output_var_names())
    ['boundary_condition_flag', 'lithosphere_surface__elevation_increment']
    >>> flexure.get_var_grid('lithosphere_surface__elevation_increment')
    0
    >>> flexure.get_grid_shape(0, np.empty(flexure.get_grid_rank(0), dtype=int))
    array([20, 40])
    >>> dz = np.empty(flexure.get_grid_size(0))
    >>> _ = flexure.get_value('lithosphere_surface__elevation_increment', dz)

    >>> np.all(dz == 0.)
    True
    >>> flexure.get_current_time()
    0.0

    >>> sorted(flexure.get_input_var_names())
    ['boundary_condition_flag', 'lithosphere__overlying_pressure_increment']
    >>> load = np.zeros((20, 40), dtype=float)
    >>> load[0, 0] = 1.
    >>> flexure.set_value('lithosphere__overlying_pressure_increment', load)
    >>> flexure.update()
    >>> flexure.get_current_time()
    2.0
    >>> _ = flexure.get_value('lithosphere_surface__elevation_increment', dz)
    >>> np.all(dz == 0.)
    False
    """
    if not issubclass(cls, Component):
        raise TypeError("class must inherit from Component")

    class BmiWrapper(Bmi):
        __doc__ = """
        Basic Modeling Interface for the {name} component.
        """.format(
            name=cls.__name__
        ).strip()

        _cls = cls

        def __init__(self):
            self._base = None
            self._clock = None
            super(BmiWrapper, self).__init__()

            self._input_var_names = tuple(
                set(self._cls._input_var_names) | {"boundary_condition_flag"}
            )
            self._output_var_names = tuple(
                set(self._cls._output_var_names) | {"boundary_condition_flag"}
            )
            self._var_mapping = self._cls._var_mapping.copy()
            self._var_units = self._cls._var_units.copy()
            self._var_doc = self._cls._var_doc.copy()

            self._var_mapping["boundary_condition_flag"] = "node"
            self._var_units["boundary_condition_flag"] = ""
            self._var_doc[
                "boundary_condition_flag"
            ] = "boundary condition flag of grid nodes"

        def get_component_name(self):
            """Name of the component."""
            return self._cls.name

        def get_input_var_names(self):
            """Names of the input exchange items."""
            return self._input_var_names

        def get_output_var_names(self):
            """Names of the output exchange items."""
            return self._output_var_names

        def get_current_time(self):
            """Current component time."""
            return self._clock.time

        def get_end_time(self):
            """Stop time for the component."""
            return self._clock.stop

        def get_start_time(self):
            """Start time of the component."""
            return self._clock.start

        def get_time_step(self):
            """Component time step."""
            return self._clock.step

        def get_time_units(self):
            """Time units used by the component."""
            return self._clock.units

        def initialize(self, config_file):
            """Initialize the component from a file.

            BMI-wrapped Landlab components use input files in YAML format.
            Component-specific parameters are listed at the top level,
            followed by grid and then time information. An example input
            file looks like::

                flexure:
                    eet: 15.e+3
                clock:
                    start: 0
                    stop: 100.
                    step: 2.
                grid:
                    type: raster
                    shape: [20, 40]
                    spacing: [1000., 2000.]

            In this case, a `RasterModelGrid` is created (with the given shape
            and spacing) and passed to the underlying landlab component. The
            `eet=15000.` is also given to the component but as a keyword
            parameter. The BMI clock is initialized with the given parameters.

            Parameters
            ----------
            config_file : str or file_like
                YAML-formatted input file for the component.
            """
            grid = create_grid(config_file, section="grid")

            if not grid:
                raise ValueError("no grid in config file ({0})".format(config_file))
            elif isinstance(grid, list):
                raise ValueError(
                    "multiple grids in config file ({0})".format(config_file)
                )

            params = load_params(config_file)
            params.pop("grid")
            clock_params = params.pop("clock")
            self._clock = TimeStepper(**clock_params)

            self._base = self._cls(grid, **params.pop(snake_case(cls.__name__), {}))
            self._base.grid.at_node[
                "boundary_condition_flag"
            ] = self._base.grid.status_at_node

        def update(self):
            """Update the component one time step."""
            if hasattr(self._base, "update"):
                self._base.update()
            elif hasattr(self._base, "run_one_step"):
                args = []
                for name, arg in inspect.signature(
                    self._base.run_one_step
                ).parameters.items():
                    if arg.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
                        args.append(name)

                if len(args) == 0 or "dt" not in args:
                    self._base.run_one_step()
                else:
                    self._base.run_one_step(self._clock.step)

            self._clock.advance()

        def update_frac(self, frac):
            """Update the component a fraction of a time step."""
            time_step = self.get_time_step()
            self._clock.step = time_step * frac
            self.update()
            self._clock.step = time_step

        def update_until(self, then):
            """Update the component until a given time."""
            n_steps = (then - self.get_current_time()) / self.get_time_step()
            for _ in range(int(n_steps)):
                self.update()
            self.update_frac(n_steps - int(n_steps))

        def finalize(self):
            """Clean-up the component."""
            pass

        def get_var_grid(self, name):
            """Get the grid id for a variable."""
            at = self._var_mapping[name]
            return BMI_GRID[at]

        def get_var_itemsize(self, name):
            """Get the size of elements of a variable."""
            at = self._var_mapping[name]
            return self._base.grid[at][name].itemsize

        def get_var_nbytes(self, name):
            """Get the total number of bytes used by a variable."""
            at = self._var_mapping[name]
            return self._base.grid[at][name].nbytes

        def get_var_type(self, name):
            """Get the data type for a variable."""
            at = self._var_mapping[name]
            return str(self._base.grid[at][name].dtype)

        def get_var_units(self, name):
            """Get the unit used by a variable."""
            return self._var_units[name]

        def get_value_ref(self, name):
            """Get a reference to a variable's data."""
            at = self._var_mapping[name]
            return self._base.grid[at][name]

        def get_value(self, name, dest):
            """Get a copy of a variable's data."""
            at = self._var_mapping[name]
            dest[:] = self._base.grid[at][name]
            return dest

        def set_value(self, name, values):
            """Set the values of a variable."""
            if name in self.get_input_var_names():
                if name == "boundary_condition_flag":
                    self._base.grid.status_at_node = values
                else:
                    at = self._var_mapping[name]
                    self._base.grid[at][name][:] = values.flat
            else:
                raise KeyError("{name} is not an input item".format(name=name))

        def get_grid_origin(self, grid, origin):
            """Get the origin for a structured grid."""
            if grid == 0:
                origin[:] = (self._base.grid.node_y[0], self._base.grid.node_x[0])
            elif grid == 1:
                origin[:] = (
                    self._base.grid.node_y[0] + self._base.grid.dy * 0.5,
                    self._base.grid.node_x[0] + self._base.grid.dx * 0.5,
                )
            return origin

        def get_grid_rank(self, grid):
            """Get the number of dimensions of a grid."""
            if grid in (0, 1):
                return 2
            else:
                return 0

        def get_grid_shape(self, grid, shape):
            """Get the shape of a structured grid."""
            if grid == 0:
                shape[:] = (
                    self._base.grid.number_of_node_rows,
                    self._base.grid.number_of_node_columns,
                )
            elif grid == 1:
                shape[:] = (
                    self._base.grid.number_of_node_rows - 1,
                    self._base.grid.number_of_node_columns - 1,
                )
            return shape

        def get_grid_spacing(self, grid, spacing):
            """Get the row and column spacing of a structured grid."""
            spacing[:] = (self._base.grid.dy, self._base.grid.dx)
            return spacing

        def get_grid_type(self, grid):
            """Get the type of grid."""
            if grid == 2:
                return "scalar"
            elif isinstance(self._base.grid, RasterModelGrid):
                return "uniform_rectilinear"
            else:
                return "unstructured"

        def get_grid_edge_count(self, grid):
            if grid == 0:
                return self._base.grid.number_of_links
            elif grid == 1:
                return self._base.grid.number_of_faces

        def get_grid_edge_nodes(self, grid, edge_nodes):
            if grid == 0:
                return self._base.grid.nodes_at_link.reshape((-1,))
            elif grid == 1:
                return self._base.grid.corners_at_face.reshape((-1,))

        def get_grid_face_count(self, grid):
            if grid == 0:
                return self._base.grid.number_of_patches
            elif grid == 1:
                return self._base.grid.number_of_cells

        def get_grid_face_nodes(self, grid, face_nodes):
            if grid == 0:
                return self._base.grid.nodes_at_patch
            elif grid == 1:
                return self._base.grid.corners_at_cell

        def get_grid_node_count(self, grid):
            if grid == 0:
                return self._base.grid.number_of_nodes
            elif grid == 1:
                return self._base.grid.number_of_corners

        def get_grid_nodes_per_face(self, grid, nodes_per_face):
            if grid == 0:
                return np.full(self._base.grid.number_of_nodes, 3, dtype=int)
            elif grid == 1:
                if isinstance(self._base.grid, HexModelGrid):
                    return np.full(self._base.grid.number_of_faces, 6, dtype=int)

        def get_grid_size(self, grid):
            if grid == 0:
                return self._base.grid.number_of_nodes
            elif grid == 1:
                return self._base.grid.number_of_corners

        def get_grid_x(self, grid, x):
            if grid == 0:
                return self._base.grid.x_of_node
            elif grid == 1:
                return self._base.grid.x_of_corner

        def get_grid_y(self, grid, y):
            if grid == 0:
                return self._base.grid.y_of_node
            elif grid == 1:
                return self._base.grid.y_of_corner

        def get_grid_z(self, grid, z):
            raise NotImplementedError("get_grid_z")
            # Only should be implemented for presently non-existant 3D grids.

        def get_value_at_indices(self, name, dest, inds):
            at = self._var_mapping[name]
            dest[:] = self._base.grid[at][name][inds]
            return dest

        def get_value_ptr(self, name):
            at = self._var_mapping[name]
            return self._base.grid[at][name]

        def get_var_location(self, name):
            return BMI_LOCATION[self._var_mapping[name]]

        def set_value_at_indices(self, name, inds, src):
            at = self._var_mapping[name]
            self._base.grid[at][name][inds] = src

    BmiWrapper.__name__ = cls.__name__
    return BmiWrapper



#! /usr/env/python

"""
Create 2D grid with randomly generated fractures.


Created: September 2013 by Greg Tucker
Last significant modification: conversion to proper component 7/2019 GT
"""

import numpy as np

from landlab import Component


def _calc_fracture_starting_position(shape, seed):
    """Choose a random starting position along the x or y axis (random choice).

    Parameters
    ----------
    shape : tuple of int
        Number of rows and columns in the grid
    seed : int
        Seeds the random number generator, so that a particular random
        sequence can be recreated.

    Returns
    -------
    (y, x) : tuple of int
        Fracture starting coordinates
    """
    np.random.seed(seed)

    if np.random.randint(0, 1) == 0:
        x = 0
        y = np.random.randint(0, shape[0] - 1)
    else:
        x = np.random.randint(0, shape[1] - 1)
        y = 0
    return (y, x)


def _calc_fracture_orientation(coords, seed):
    """Choose a random orientation for the fracture.

    Parameters
    ----------
    coords : tuple of int
        Starting coordinates (one of which should be zero) as *y*, *x*.
    seed : int
        Seed value for random number generator

    Returns
    -------
    ang : float
        Fracture angle relative to horizontal

    Notes
    -----
    If the fracture starts along the bottom of the grid (y=0), then the
    angle will be between 45 and 135 degrees from horizontal
    (counter-clockwise). Otherwise, it will be between -45 and 45 degrees.
    """
    y, x = coords

    np.random.seed(seed)
    ang = (np.pi / 2) * np.random.rand()
    if y == 0:
        ang += np.pi / 4
    else:
        ang -= np.pi / 4

    return ang


def _calc_fracture_step_sizes(start_yx, ang):
    """
    Calculate the sizes of steps dx and dy to be used when "drawing" the
    fracture onto the grid.

    Parameters
    ----------
    start_yx : tuple of int
        Starting grid coordinates
    ang : float
        Fracture angle relative to horizontal (radians)

    Returns
    -------
    (dy, dx) : tuple of float
        Step sizes in y and x directions. One will always be unity, and the
        other will always be <1.
    """
    starty, startx = start_yx
    if startx == 0:  # frac starts on left side
        dx = 1
        dy = np.tan(ang)
    else:  # frac starts on bottom side
        dy = 1
        dx = -np.tan(ang - np.pi / 2)

    return (dy, dx)


def _trace_fracture_through_grid(m, start_yx, spacing):
    """Create a 2D fracture in a grid.

    Creates a "fracture" in a 2D grid, m, by setting cell values to unity
    along the trace of the fracture (i.e., "drawing" a line throuh the
    grid).

    Parameters
    ----------
    m : 2D Numpy array
        Array that represents the grid
    start_yx : tuple of int
        Starting grid coordinates for fracture
    spacing : tuple of float
        Step sizes in y and x directions

    Returns
    -------
    None, but changes contents of m
    """
    y0, x0 = start_yx
    dy, dx = spacing

    x = x0
    y = y0

    while (
        round(x) < np.size(m, 1)
        and round(y) < np.size(m, 0)
        and round(x) >= 0
        and round(y) >= 0
    ):
        m[int(y + 0.5)][int(x + 0.5)] = 1
        x += dx
        y += dy


class FractureGridGenerator(Component):

    """
    Create a 2D grid with randomly generated fractures.

    The grid contains the value 1 where fractures (one cell wide) exist, and
    0 elsewhere. The idea is to use this for simulations based on weathering
    and erosion of, and/or flow within, fracture networks.

    Parameters
    ----------
    frac_spacing : int, optional
        Average spacing of fractures (in grid cells) (default = 10)
    seed : int, optional
        Seed used for random number generator (default = 0)

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> grid = RasterModelGrid((5, 5))
    >>> fg = FractureGridGenerator(grid=grid, frac_spacing=3)
    >>> fg.run_one_step()
    >>> grid.at_node['fracture_at_node']
    array([1, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
           0, 0], dtype=int8)

    Notes
    -----
    Potential improvements:
    - Add doctests
    - Fractures could be defined by links rather than nodes (i.e., return a
        link array with a code indicating whether the link crosses a fracture
        or not)
    - Fractures could have a finite length rather than extending all the way
        across the grid
    - Use of starting position along either x or y axis makes fracture net
        somewhat asymmetric. One would need a different algorithm to make it
        fully (statistically) symmetric.
    """

    _name = "FractureGridGenerator"

    _input_var_names = ()

    _output_var_names = ("fracture_at_node",)

    _var_units = {"fracture_at_node": "-"}

    _var_mapping = {"fracture_at_node": "node"}

    _var_doc = {"fracture_at_node": "presence (1) or absence (0) of fracture"}

    def __init__(self, grid, frac_spacing=10.0, seed=0):
        """Initialize the FractureGridGenerator."""

        self._grid = grid
        self.frac_spacing = frac_spacing
        self.seed = seed
        super(FractureGridGenerator, self).__init__(grid)

        # TODO: delete this once we have generation of output fields in
        # base class
        if "fracture_at_node" not in grid.at_node:
            grid.add_zeros("node", "fracture_at_node", dtype=np.int8)

    def run_one_step(self):
        """Run FractureGridGenerator and create a random fracture grid."""
        self._make_frac_grid(self.frac_spacing, self.seed)

    def _make_frac_grid(self, frac_spacing, seed):
        """Create a grid that contains a network of random fractures.

        Creates a grid containing a network of random fractures, which are
        represented as 1's embedded in a grid of 0's. The grid is stored in
        the "fracture_at_node" field.

        Parameters
        ----------
        frac_spacing : int
            Average spacing of fractures (in grid cells)
        seed : int
            Seed used for random number generator
        """
        # Make an initial grid of all zeros. If user specified a model grid,
        # use that. Otherwise, use the given dimensions.
        nr = self._grid.number_of_node_rows
        nc = self._grid.number_of_node_columns
        m = self._grid.at_node["fracture_at_node"].reshape((nr, nc))

        # Add fractures to grid
        nfracs = (nr + nc) // frac_spacing
        for i in range(nfracs):

            (y, x) = _calc_fracture_starting_position((nr, nc), seed + i)
            ang = _calc_fracture_orientation((y, x), seed + i)
            (dy, dx) = _calc_fracture_step_sizes((y, x), ang)

            _trace_fracture_through_grid(m, (y, x), (dy, dx))



from functools import wraps

import six


def track_this_method(func):
    @wraps(func)
    def _wrapped(self, *args, **kwds):
        if self._DEBUG_TRACK_METHODS:
            six.print_("Entering: %s.%s" % (self.__class__.__name__, func.__name__))
        ans = func(self, *args, **kwds)
        if self._DEBUG_TRACK_METHODS:
            six.print_("Exiting: %s.%s" % (self.__class__.__name__, func.__name__))
        return ans

    return _wrapped



"""
Unit tests for \
    landlab.components.soil_moisture.soil_moisture_dynamics
"""
import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal

(_SHAPE, _SPACING, _ORIGIN) = ((20, 20), (10e0, 10e0), (0.0, 0.0))
_ARGS = (_SHAPE, _SPACING, _ORIGIN)


def test_name(sm):
    assert sm.name == "Soil Moisture"


def test_input_var_names(sm):
    assert sorted(sm.input_var_names) == [
        "rainfall__daily_depth",
        "soil_moisture__initial_saturation_fraction",
        "surface__potential_evapotranspiration_rate",
        "vegetation__cover_fraction",
        "vegetation__live_leaf_area_index",
        "vegetation__plant_functional_type",
    ]


def test_output_var_names(sm):
    assert sorted(sm.output_var_names) == [
        "soil_moisture__root_zone_leakage",
        "soil_moisture__saturation_fraction",
        "surface__evapotranspiration",
        "surface__runoff",
        "vegetation__water_stress",
    ]


def test_var_units(sm):
    assert set(sm.input_var_names) | set(sm.output_var_names), set(
        dict(sm.units).keys()
    )

    assert sm.var_units("vegetation__cover_fraction") == "None"
    assert sm.var_units("vegetation__live_leaf_area_index") == "None"
    assert sm.var_units("surface__potential_evapotranspiration_rate") == "mm"
    assert sm.var_units("vegetation__plant_functional_type") == "None"
    assert sm.var_units("vegetation__water_stress") == "None"
    assert sm.var_units("soil_moisture__saturation_fraction") == "None"
    assert sm.var_units("soil_moisture__initial_saturation_fraction") == "None"
    assert sm.var_units("soil_moisture__root_zone_leakage") == "mm"
    assert sm.var_units("surface__runoff") == "mm"
    assert sm.var_units("surface__evapotranspiration") == "mm"
    assert sm.var_units("rainfall__daily_depth") == "mm"


def test_grid_shape(sm):
    assert sm.grid.number_of_node_rows == _SHAPE[0]
    assert sm.grid.number_of_node_columns == _SHAPE[1]


def test_grid_x_extent(sm):
    assert sm.grid.extent[1] == (_SHAPE[1] - 1) * _SPACING[1]


def test_grid_y_extent(sm):
    assert sm.grid.extent[0] == (_SHAPE[0] - 1) * _SPACING[0]


def test_field_getters(sm):
    for name in sm.grid["node"]:
        field = sm.grid["node"][name]
        assert isinstance(field, np.ndarray)
        assert field.shape == (
            sm.grid.number_of_node_rows * sm.grid.number_of_node_columns,
        )

    for name in sm.grid["cell"]:
        field = sm.grid["cell"][name]
        assert isinstance(field, np.ndarray)
        assert field.shape == (
            sm.grid.number_of_cell_rows * sm.grid.number_of_cell_columns,
        )

    with pytest.raises(KeyError):
        sm.grid["not_a_var_name"]


def test_field_initialized_to_zero(sm):
    for name in sm.grid["node"]:
        field = sm.grid["node"][name]
        assert_array_almost_equal(field, np.zeros(sm.grid.number_of_nodes))
    for name in sm.grid["cell"]:
        field = sm.grid["cell"][name]
        assert_array_almost_equal(field, np.zeros(sm.grid.number_of_cells))



# -*- coding: utf-8 -*-
"""
Landlab component for overland flow using the kinematic-wave approximation.

Created on Fri May 27 14:26:13 2016

@author: gtucker
"""


import numpy as np

from landlab import Component


class KinwaveOverlandFlowModel(Component):
    """Calculate water flow over topography.

    Landlab component that implements a two-dimensional
    kinematic wave model. This is an extremely simple, unsophisticated
    model, originally built simply to demonstrate the component creation
    process. Limitations to the present version include: infiltration is
    handled very crudely, the called is responsible for picking a stable
    time step size (no adaptive time stepping is used in the `run_one_step`
    method), precipitation rate is constant for a given duration (then zero),
    and all parameters are uniform in space. Also, the terrain is assumed
    to be stable over time. Caveat emptor!

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> rg = RasterModelGrid((4, 5), xy_spacing=10.0)
    >>> kw = KinwaveOverlandFlowModel(rg)
    >>> kw.vel_coef
    100.0
    >>> rg.at_node['surface_water__depth']
    array([ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,
            0.,  0.,  0.,  0.,  0.,  0.,  0.])
    """

    _name = "KinwaveOverlandFlowModel"

    _input_var_names = ("topographic__elevation", "topographic__gradient")

    _output_var_names = (
        "surface_water__depth",
        "water__velocity",
        "water__specific_discharge",
    )

    _var_units = {
        "topographic__elevation": "m",
        "topographic__gradient": "m/m",
        "surface_water__depth": "m",
        "water__velocity": "m/s",
        "water__specific_discharge": "m2/s",
    }

    _var_mapping = {
        "topographic__elevation": "node",
        "topographic__gradient": "link",
        "surface_water__depth": "node",
        "water__velocity": "link",
        "water__specific_discharge": "link",
    }

    _var_doc = {
        "topographic__elevation": "elevation of the ground surface relative to some datum",
        "topographic__gradient": "gradient of the ground surface",
        "surface_water__depth": "depth of water",
        "water__velocity": "flow velocity component in the direction of the link",
        "water__specific_discharge": "flow discharge component in the direction of the link",
    }

    def __init__(
        self,
        grid,
        precip_rate=1.0,
        precip_duration=1.0,
        infilt_rate=0.0,
        roughness=0.01,
        **kwds
    ):
        """Initialize the KinwaveOverlandFlowModel.

        Parameters
        ----------
        grid : ModelGrid
            Landlab ModelGrid object
        precip_rate : float, optional (defaults to 1 mm/hr)
            Precipitation rate, mm/hr
        precip_duration : float, optional (defaults to 1 hour)
            Duration of precipitation, hours
        infilt_rate : float, optional (defaults to 0)
            Maximum rate of infiltration, mm/hr
        roughness : float, defaults to 0.01
            Manning roughness coefficient, s/m^1/3
        """

        # Store grid and parameters and do unit conversion
        self._grid = grid
        self.precip = precip_rate / 3600000.0  # convert to m/s
        self.precip_duration = precip_duration * 3600.0  # h->s
        self.infilt = infilt_rate / 3600000.0  # convert to m/s
        self.vel_coef = 1.0 / roughness  # do division now to save time

        # Create fields...
        #   Elevation
        if "topographic__elevation" in grid.at_node:
            self.elev = grid.at_node["topographic__elevation"]
        else:
            self.elev = grid.add_zeros("node", "topographic__elevation")
        #  Water depth
        if "surface_water__depth" in grid.at_node:
            self.depth = grid.at_node["surface_water__depth"]
        else:
            self.depth = grid.add_zeros("node", "surface_water__depth")
        #   Slope
        if "topographic__gradient" in grid.at_link:
            self.slope = grid.at_link["topographic__gradient"]
        else:
            self.slope = grid.add_zeros("link", "topographic__gradient")
        #  Velocity
        if "water__velocity" in grid.at_link:
            self.vel = grid.at_link["water__velocity"]
        else:
            self.vel = grid.add_zeros("link", "water__velocity")
        #  Discharge
        if "water__specific_discharge" in grid.at_link:
            self.disch = grid.at_link["water__specific_discharge"]
        else:
            self.disch = grid.add_zeros("link", "water__specific_discharge")

        # Calculate the ground-surface slope (assume it won't change)
        self.slope[self._grid.active_links] = self._grid.calc_grad_at_link(self.elev)[
            self._grid.active_links
        ]
        self.sqrt_slope = np.sqrt(self.slope)
        self.sign_slope = np.sign(self.slope)

    def run_one_step(self, dt, current_time=0.0, **kwds):
        """Calculate water flow for a time period `dt`.

        Default units for dt are *seconds*.
        """
        # Calculate water depth at links. This implements an "upwind" scheme
        # in which water depth at the links is the depth at the higher of the
        # two nodes.
        H_link = self._grid.map_value_at_max_node_to_link(
            "topographic__elevation", "surface_water__depth"
        )

        # Calculate velocity using the Manning equation.
        self.vel = (
            -self.sign_slope * self.vel_coef * H_link ** 0.66667 * self.sqrt_slope
        )

        # Calculate discharge
        self.disch = H_link * self.vel

        # Flux divergence
        dqda = self._grid.calc_flux_div_at_node(self.disch)

        # Rate of change of water depth
        if current_time < self.precip_duration:
            ppt = self.precip
        else:
            ppt = 0.0
        dHdt = ppt - self.infilt - dqda

        # Update water depth: simple forward Euler scheme
        self.depth[self._grid.core_nodes] += dHdt[self._grid.core_nodes] * dt

        # Very crude numerical hack: prevent negative water depth
        self.depth[np.where(self.depth < 0.0)[0]] = 0.0


if __name__ == "__main__":
    import doctest

    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Unit tests for landlab.components.overland_flow.KinwaveOverlandFlowModel

last updated: 3/14/16
"""
(_SHAPE, _SPACING, _ORIGIN) = ((10, 10), (25, 25), (0.0, 0.0))
_ARGS = (_SHAPE, _SPACING, _ORIGIN)


def test_KinWaveOF_name(kin_wave_of):
    assert kin_wave_of.name == "KinwaveOverlandFlowModel"


def test_KinWaveOF_input_var_names(kin_wave_of):
    assert kin_wave_of.input_var_names == (
        "topographic__elevation",
        "topographic__gradient",
    )


def test_KinWaveOF_output_var_names(kin_wave_of):
    assert kin_wave_of.output_var_names == (
        "surface_water__depth",
        "water__velocity",
        "water__specific_discharge",
    )


def test_KinWaveOF_var_units(kin_wave_of):
    assert set(kin_wave_of.input_var_names) | set(kin_wave_of.output_var_names) == set(
        dict(kin_wave_of.units).keys()
    )

    assert kin_wave_of.var_units("topographic__elevation") == "m"
    assert kin_wave_of.var_units("topographic__gradient") == "m/m"
    assert kin_wave_of.var_units("surface_water__depth") == "m"
    assert kin_wave_of.var_units("water__velocity") == "m/s"
    assert kin_wave_of.var_units("water__specific_discharge") == "m2/s"


def test_grid_shape(kin_wave_of):
    assert kin_wave_of.grid.number_of_node_rows == _SHAPE[0]
    assert kin_wave_of.grid.number_of_node_columns == _SHAPE[1]


def test_run_one_step():
    from landlab import RasterModelGrid
    import numpy as np
    from landlab.components.overland_flow import KinwaveOverlandFlowModel

    grid = RasterModelGrid((10, 10), xy_spacing=0.5)
    grid.add_zeros("node", "topographic__elevation", dtype=float)
    grid.add_zeros("node", "topographic__gradient")

    topo_arr = np.ones(100).reshape(10, 10)
    i = 0
    while i <= 9:
        topo_arr[:, i] = 5 + (0.002 * i)
        i += 1
    topo_arr = topo_arr.flatten()
    grid["node"]["topographic__elevation"] = topo_arr
    KinWaveOF = KinwaveOverlandFlowModel(
        grid, precip_rate=100.0, precip_duration=1.0, roughness=0.02
    )

    KinWaveOF.run_one_step(60)

    # I'll admit this is very non-robust. Solution roughly based on plot #9
    # from Heng et. al, (2009): "Modeling overland flow and soil eroion on
    # non uniform hillslopes: A finite volume scheme." They do not provide the
    # numerical solution but the plots match...
    max_h_mm = max(grid["node"]["surface_water__depth"]) * 1000.0
    np.testing.assert_almost_equal(max_h_mm, 1.66666666667)



from six import StringIO

from landlab.core import load_params

YAML_PARAMS_STR = """
x: 1e7
y: 1
z: [1, 2]
a: frog
"""
MPD_PARAMS_STR = """
x: A float
1e7
y: An int
1
a: A string
frog
"""
YAML_PARAMS = {"x": 1e7, "y": 1, "z": [1, 2], "a": "frog"}
MPD_PARAMS = {"x": 1e7, "y": 1, "a": "frog"}


def test_from_yaml_string():
    """Load parameters from YAML-formatted string."""
    params = load_params(YAML_PARAMS_STR)
    assert params == YAML_PARAMS
    assert isinstance(params["x"], float)
    assert isinstance(params["y"], int)


def test_from_yaml_file_like():
    """Load parameters from YAML-formatted string."""
    params = load_params(StringIO(YAML_PARAMS_STR))
    assert params == YAML_PARAMS
    assert isinstance(params["x"], float)
    assert isinstance(params["y"], int)


def test_from_yaml_path(tmpdir):
    """Load parameters from YAML-formatted string."""
    with tmpdir.as_cwd():
        with open("params.yaml", "w") as fp:
            fp.write(YAML_PARAMS_STR)
        params = load_params("./params.yaml")
    assert params == YAML_PARAMS
    assert isinstance(params["x"], float)
    assert isinstance(params["y"], int)


def test_from_mpd_string():
    """Load parameters from YAML-formatted string."""
    params = load_params(MPD_PARAMS_STR)
    assert params == MPD_PARAMS
    assert isinstance(params["x"], float)
    assert isinstance(params["y"], int)


def test_from_mpd_file_like():
    """Load parameters from YAML-formatted string."""
    params = load_params(StringIO(MPD_PARAMS_STR))
    assert params == MPD_PARAMS
    assert isinstance(params["x"], float)
    assert isinstance(params["y"], int)


def test_from_mpd_path(tmpdir):
    """Load parameters from YAML-formatted string."""
    with tmpdir.as_cwd():
        with open("params.txt", "w") as fp:
            fp.write(MPD_PARAMS_STR)
        params = load_params("./params.txt")
    assert params == MPD_PARAMS
    assert isinstance(params["x"], float)
    assert isinstance(params["y"], int)



# -*- coding: utf-8 -*-
"""
Unit tests for landlab.components.overland_flow.OverlandFlow

last updated: 3/14/16
"""
import numpy as np

from landlab import RasterModelGrid
from landlab.components.overland_flow import OverlandFlow
from landlab.grid.structured_quad.links import left_edge_horizontal_ids

(_SHAPE, _SPACING, _ORIGIN) = ((32, 240), (25, 25), (0.0, 0.0))
_ARGS = (_SHAPE, _SPACING, _ORIGIN)


def test_deAlm_name(deAlm):
    assert deAlm.name == "OverlandFlow"


def test_deAlm_input_var_names(deAlm):
    assert deAlm.input_var_names == ("surface_water__depth", "topographic__elevation")


def test_deAlm_output_var_names(deAlm):
    assert deAlm.output_var_names == (
        "surface_water__depth",
        "surface_water__discharge",
        "water_surface__gradient",
    )


def test_deAlm_var_units(deAlm):
    assert set(deAlm.input_var_names) | set(deAlm.output_var_names) == set(
        dict(deAlm.units).keys()
    )

    assert deAlm.var_units("surface_water__depth") == "m"
    assert deAlm.var_units("surface_water__discharge") == "m3/s"
    assert deAlm.var_units("water_surface__gradient") == "-"
    assert deAlm.var_units("topographic__elevation") == "m"


def test_grid_shape(deAlm):
    assert deAlm.grid.number_of_node_rows == _SHAPE[0]
    assert deAlm.grid.number_of_node_columns == _SHAPE[1]


def test_deAlm_analytical():
    grid = RasterModelGrid((32, 240), xy_spacing=25)
    grid.add_zeros("node", "surface_water__depth")
    grid.add_zeros("node", "topographic__elevation")
    grid.set_closed_boundaries_at_grid_edges(True, True, True, True)
    left_inactive_ids = left_edge_horizontal_ids(grid.shape)
    deAlm = OverlandFlow(grid, mannings_n=0.01, h_init=0.001)
    time = 0.0

    while time < 500.0:
        grid.at_link["surface_water__discharge"][left_inactive_ids] = grid.at_link[
            "surface_water__discharge"
        ][left_inactive_ids + 1]
        dt = deAlm.calc_time_step()
        deAlm.overland_flow(dt)
        h_boundary = ((7.0 / 3.0) * (0.01 ** 2) * (0.4 ** 3) * time) ** (3.0 / 7.0)
        grid.at_node["surface_water__depth"][grid.nodes[1:-1, 1]] = h_boundary
        time += dt

    x = np.arange(0, ((grid.shape[1]) * grid.dx), grid.dx)
    h_analytical = -(7.0 / 3.0) * (0.01 ** 2) * (0.4 ** 2) * (x - (0.4 * 500))

    h_analytical[np.where(h_analytical > 0)] = h_analytical[
        np.where(h_analytical > 0)
    ] ** (3.0 / 7.0)
    h_analytical[np.where(h_analytical < 0)] = 0.0

    hdeAlm = deAlm.h.reshape(grid.shape)
    hdeAlm = hdeAlm[1][1:]
    hdeAlm = np.append(hdeAlm, [0])
    np.testing.assert_almost_equal(h_analytical, hdeAlm, decimal=1)


def test_deAlm_analytical_imposed_dt_short():
    grid = RasterModelGrid((32, 240), xy_spacing=25)
    grid.add_zeros("node", "surface_water__depth")
    grid.add_zeros("node", "topographic__elevation")
    grid.set_closed_boundaries_at_grid_edges(True, True, True, True)
    left_inactive_ids = left_edge_horizontal_ids(grid.shape)
    deAlm = OverlandFlow(grid, mannings_n=0.01, h_init=0.001)
    time = 0.0

    while time < 500.0:
        grid.at_link["surface_water__discharge"][left_inactive_ids] = grid.at_link[
            "surface_water__discharge"
        ][left_inactive_ids + 1]
        dt = 10.0
        deAlm.overland_flow(dt)
        h_boundary = ((7.0 / 3.0) * (0.01 ** 2) * (0.4 ** 3) * time) ** (3.0 / 7.0)
        grid.at_node["surface_water__depth"][grid.nodes[1:-1, 1]] = h_boundary
        time += dt

    x = np.arange(0, ((grid.shape[1]) * grid.dx), grid.dx)
    h_analytical = -(7.0 / 3.0) * (0.01 ** 2) * (0.4 ** 2) * (x - (0.4 * 500))

    h_analytical[np.where(h_analytical > 0)] = h_analytical[
        np.where(h_analytical > 0)
    ] ** (3.0 / 7.0)
    h_analytical[np.where(h_analytical < 0)] = 0.0

    hdeAlm = deAlm.h.reshape(grid.shape)
    hdeAlm = hdeAlm[1][1:]
    hdeAlm = np.append(hdeAlm, [0])
    np.testing.assert_almost_equal(h_analytical, hdeAlm, decimal=1)


def test_deAlm_analytical_imposed_dt_long():
    grid = RasterModelGrid((32, 240), xy_spacing=25)
    grid.add_zeros("node", "surface_water__depth")
    grid.add_zeros("node", "topographic__elevation")
    grid.set_closed_boundaries_at_grid_edges(True, True, True, True)
    left_inactive_ids = left_edge_horizontal_ids(grid.shape)
    deAlm = OverlandFlow(grid, mannings_n=0.01, h_init=0.001)
    time = 0.0

    while time < 500.0:
        grid.at_link["surface_water__discharge"][left_inactive_ids] = grid.at_link[
            "surface_water__discharge"
        ][left_inactive_ids + 1]
        dt = 100.0
        deAlm.run_one_step(dt)
        h_boundary = ((7.0 / 3.0) * (0.01 ** 2) * (0.4 ** 3) * time) ** (3.0 / 7.0)
        grid.at_node["surface_water__depth"][grid.nodes[1:-1, 1]] = h_boundary
        time += dt

    x = np.arange(0, ((grid.shape[1]) * grid.dx), grid.dx)
    h_analytical = -(7.0 / 3.0) * (0.01 ** 2) * (0.4 ** 2) * (x - (0.4 * 500))

    h_analytical[np.where(h_analytical > 0)] = h_analytical[
        np.where(h_analytical > 0)
    ] ** (3.0 / 7.0)
    h_analytical[np.where(h_analytical < 0)] = 0.0

    hdeAlm = deAlm.h.reshape(grid.shape)
    hdeAlm = hdeAlm[1][1:]
    hdeAlm = np.append(hdeAlm, [0])
    np.testing.assert_almost_equal(h_analytical, hdeAlm, decimal=1)



#! /usr/bin/env python
"""
A class used to create and manage network models in 2D.
"""
import numpy as np

from landlab.grid.base import _default_axis_names, _default_axis_units
from landlab.utils.decorators import make_return_array_immutable

from ..core import load_params
from ..core.utils import add_module_functions_to_class
from ..field import GraphFields
from ..graph import NetworkGraph
from ..utils.decorators import cache_result_in_object
from .decorators import override_array_setitem_and_reset, return_readonly_id_array
from .linkstatus import ACTIVE_LINK, set_status_at_link


class NetworkModelGrid(NetworkGraph, GraphFields):
    """Create a ModelGrid of just nodes and links.

    Parameters
    ----------
    yx_of_node : tuple of ndarray
        Node y and x coordinates.
    links : array of tuple of int
        Nodes at link tail and head.
    xy_of_reference : tuple, optional
        Coordinate value in projected space of (0., 0.)
        Default is (0., 0.)

    Examples
    --------
    >>> from landlab import NetworkModelGrid
    >>> y_of_node = (0, 1, 2, 2)
    >>> x_of_node = (0, 0, -1, 1)
    >>> nodes_at_link = ((1, 0), (2, 1), (3, 1))
    >>> grid = NetworkModelGrid((y_of_node, x_of_node), nodes_at_link)
    >>> grid.x_of_node
    array([ 0.,  0., -1.,  1.])
    >>> grid.y_of_node
    array([ 0.,  1.,  2.,  2.])
    >>> grid.nodes_at_link
    array([[0, 1],
           [2, 1],
           [1, 3]])
    """

    def __init__(self, yx_of_node, links, **kwds):
        NetworkGraph.__init__(self, yx_of_node, links=links)
        GraphFields.__init__(
            self,
            {"node": self.number_of_nodes, "link": self.number_of_links, "grid": 1},
            default_group="node",
        )

        self._node_status = np.zeros(self.number_of_nodes, dtype=np.uint8)
        self.bc_set_code = 0

        self.axis_name = kwds.get("axis_name", _default_axis_names(self.ndim))

        self.axis_units = kwds.get("axis_units", _default_axis_units(self.ndim))

        self._ref_coord = tuple(kwds.get("xy_of_reference", (0.0, 0.0)))

    @classmethod
    def from_file(cls, file_like):
        params = load_params(file_like)
        return cls.from_dict(params)

    @classmethod
    def from_dict(cls, params):
        return cls(**params)

    @property
    def xy_of_reference(self):
        """Return the coordinates (x, y) of the reference point.

        For RasterModelGrid and HexModelGrid the reference point is the
        minimum of x_of_node and of y_of_node. By default it is (0, 0). For
        VoronoiDelaunayGrid the reference point is (0, 0). For RadialModelGrid
        it is the (x, y) of the center point.

        The intention of these coordinates is to provide a method to store
        the large float values of projected coordinates.

        Example
        -------

        >>> from landlab import NetworkModelGrid
        >>> y_of_node = (0, 1, 2, 2)
        >>> x_of_node = (0, 0, -1, 1)
        >>> nodes_at_link = ((1, 0), (2, 1), (3, 1))
        >>> grid = NetworkModelGrid((y_of_node, x_of_node),
        ...                         nodes_at_link,
        ...                         xy_of_reference = (12345, 678910))
        >>> grid.xy_of_reference
        (12345, 678910)
        >>> grid.xy_of_reference = (98765, 43210)
        >>> grid.xy_of_reference
        (98765, 43210)
        """
        return self._ref_coord

    @xy_of_reference.setter
    def xy_of_reference(self, new_xy_of_reference):
        """Set a new value for the model grid xy_of_reference."""
        self._ref_coord = (new_xy_of_reference[0], new_xy_of_reference[1])

    @property
    def axis_units(self):
        """Get units for each axis.

        Returns
        -------
        tuple of str
            The units (as a string) for each of a grid's coordinates.

        Examples
        --------
        >>> from landlab import NetworkModelGrid
        >>> y_of_node = (0, 1, 2, 2)
        >>> x_of_node = (0, 0, -1, 1)
        >>> nodes_at_link = ((1, 0), (2, 1), (3, 1))
        >>> grid = NetworkModelGrid((y_of_node, x_of_node),
        ...                         nodes_at_link)
        >>> grid.axis_units
        ('-', '-')
        >>> grid.axis_units = ('km', 'km')
        >>> grid.axis_units
        ('km', 'km')

        LLCATS: GINF
        """
        return self._axis_units

    @axis_units.setter
    def axis_units(self, new_units):
        """Set the units for each coordinate axis."""
        if len(new_units) != self.ndim:
            raise ValueError("length of units does not match grid dimension")
        self._axis_units = tuple(new_units)

    @property
    def axis_name(self):
        """Get the name of each coordinate axis.

        Returns
        -------
        tuple of str
            The names of each axis.

        Examples
        --------
        >>> from landlab import NetworkModelGrid
        >>> y_of_node = (0, 1, 2, 2)
        >>> x_of_node = (0, 0, -1, 1)
        >>> nodes_at_link = ((1, 0), (2, 1), (3, 1))
        >>> grid = NetworkModelGrid((y_of_node, x_of_node),
        ...                         nodes_at_link)
        >>> grid.axis_name
        ('y', 'x')
        >>> grid.axis_name = ('lon', 'lat')
        >>> grid.axis_name
        ('lon', 'lat')

        LLCATS: GINF
        """
        return self._axis_name

    @axis_name.setter
    def axis_name(self, new_names):
        """Set the names of a grid's coordinate axes.

        Raises
        ------
        ValueError
            If the number of dimension do not match.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> grid = RasterModelGrid((4, 5))
        >>> grid.axis_name = ('lon', 'lat')
        >>> grid.axis_name
        ('lon', 'lat')
        """
        if len(new_names) != self.ndim:
            raise ValueError("length of names does not match grid dimension")
        self._axis_name = tuple(new_names)

    @property
    @override_array_setitem_and_reset("reset_status_at_node")
    def status_at_node(self):
        """Get array of the boundary status for each node.

        Examples
        --------
        >>> from landlab import NetworkModelGrid
        >>> from landlab import CLOSED_BOUNDARY, CORE_NODE

        >>> y_of_node = (0, 1, 2, 2)
        >>> x_of_node = (0, 0, -1, 1)
        >>> nodes_at_link = ((1, 0), (2, 1), (3, 1))
        >>> grid = NetworkModelGrid((y_of_node, x_of_node), nodes_at_link)
        >>> grid.status_at_node
        array([0, 0, 0, 0], dtype=uint8)
        >>> grid.status_at_link
        array([0, 0, 0], dtype=uint8)

        Now we change the status at node 0 to a closed boundary. This will
        result in changing the link status as well.

        >>> grid.status_at_node = [CLOSED_BOUNDARY, CORE_NODE, CORE_NODE, CORE_NODE]
        >>> grid.status_at_node
        array([4, 0, 0, 0], dtype=uint8)
        >>> grid.status_at_link
        array([4, 0, 0], dtype=uint8)

        LLCATS: NINF BC
        """
        return self._node_status

    @status_at_node.setter
    def status_at_node(self, new_status):
        """Set the array of node boundary statuses."""
        self._node_status[:] = new_status[:]
        self.reset_status_at_node()

    def reset_status_at_node(self):
        attrs = [
            "_active_link_dirs_at_node",
            "_status_at_link",
            "_active_links",
            "_fixed_links",
            "_activelink_fromnode",
            "_activelink_tonode",
            "_fixed_links",
            "_active_adjacent_nodes_at_node",
            "_fixed_value_boundary_nodes",
            "_link_status_at_node",
        ]
        for attr in attrs:
            try:
                del self.__dict__[attr]
            except KeyError:
                pass

        self.bc_set_code += 1

    @property
    @make_return_array_immutable
    @cache_result_in_object()
    def status_at_link(self):
        """Get array of the status of all links.

        Examples
        --------
        >>> from landlab import NetworkModelGrid
        >>> y_of_node = (0, 1, 2, 2)
        >>> x_of_node = (0, 0, -1, 1)
        >>> nodes_at_link = ((1, 0), (2, 1), (3, 1))
        >>> grid = NetworkModelGrid((y_of_node, x_of_node), nodes_at_link)
        >>> grid.status_at_link
        array([0, 0, 0], dtype=uint8)

        LLCATS: LINF BC
        """
        return set_status_at_link(self.status_at_node[self.nodes_at_link])

    @property
    @return_readonly_id_array
    @cache_result_in_object()
    def active_links(self):
        """Get array of active links.

        Examples
        --------
        >>> from landlab import NetworkModelGrid
        >>> from landlab import FIXED_LINK

        >>> y_of_node = (0, 1, 2, 2)
        >>> x_of_node = (0, 0, -1, 1)
        >>> nodes_at_link = ((1, 0), (2, 1), (3, 1))
        >>> grid = NetworkModelGrid((y_of_node, x_of_node), nodes_at_link)
        >>> grid.active_links
        array([0, 1, 2])

        LLCATS: NINF BC SUBSET
        """
        return np.where(self.status_at_link == ACTIVE_LINK)[0]

    @property
    @cache_result_in_object()
    @make_return_array_immutable
    def x_of_link(self):
        """Get array of the x-coordinates of link midpoints.

        Examples
        --------
        >>> from landlab import NetworkModelGrid
        >>> y_of_node = (0, 1, 2, 2)
        >>> x_of_node = (0, 0, -1, 1)
        >>> nodes_at_link = ((1, 0), (2, 1), (3, 1))
        >>> grid = NetworkModelGrid((y_of_node, x_of_node), nodes_at_link)
        >>> grid.x_of_link
        array([ 0. , -0.5,  0.5])

        LLCATS: LINF MEAS
        """
        return np.mean(self.x_of_node[self.nodes_at_link], axis=1)

    @property
    @cache_result_in_object()
    @make_return_array_immutable
    def y_of_link(self):
        """Get array of the y-coordinates of link midpoints.

        Examples
        --------
        >>> from landlab import NetworkModelGrid
        >>> y_of_node = (0, 1, 2, 2)
        >>> x_of_node = (0, 0, -1, 1)
        >>> nodes_at_link = ((1, 0), (2, 1), (3, 1))
        >>> grid = NetworkModelGrid((y_of_node, x_of_node), nodes_at_link)
        >>> grid.y_of_link
        array([ 0.5,  1.5,  1.5])

        LLCATS: LINF MEAS
        """
        return np.mean(self.y_of_node[self.nodes_at_link], axis=1)


# add only the correct functions
add_module_functions_to_class(
    NetworkModelGrid, "mappers.py", pattern="map_*", exclude="cell|patch"
)
add_module_functions_to_class(
    NetworkModelGrid, "gradients.py", pattern="calc_grad_at_link"
)



from .dual import DualGraph
from .graph import Graph, NetworkGraph
from .hex import DualHexGraph, HexGraph
from .radial import DualRadialGraph, RadialGraph
from .structured_quad import (
    DualRectilinearGraph,
    DualStructuredQuadGraph,
    DualUniformRectilinearGraph,
    RectilinearGraph,
    StructuredQuadGraph,
    UniformRectilinearGraph,
)
from .voronoi import DualVoronoiGraph, VoronoiGraph

__all__ = [
    "Graph",
    "NetworkGraph",
    "DualGraph",
    "StructuredQuadGraph",
    "RectilinearGraph",
    "UniformRectilinearGraph",
    "DualUniformRectilinearGraph",
    "DualRectilinearGraph",
    "DualStructuredQuadGraph",
    "VoronoiGraph",
    "DualVoronoiGraph",
    "HexGraph",
    "DualHexGraph",
    "RadialGraph",
    "DualRadialGraph",
]



#!/usr/env/python
import time

from matplotlib.pyplot import axis
from numpy import random

from landlab.ca.celllab_cts import CAPlotter, Transition
from landlab.io.native_landlab import save_grid

_DEBUG = False


class CTSModel(object):
    """
    Implement a generic CellLab-CTS model.

    This is the base class from which models should inherit.
    """

    def __init__(
        self,
        grid_size=(5, 5),
        report_interval=5.0,
        grid_orientation="vertical",
        node_layout="rect",
        show_plots=False,
        cts_type="oriented_hex",
        run_duration=1.0,
        output_interval=1.0e99,
        plot_every_transition=False,
        **kwds
    ):

        self.initialize(
            grid_size,
            report_interval,
            grid_orientation,
            node_layout,
            show_plots,
            cts_type,
            run_duration,
            output_interval,
            plot_every_transition,
            **kwds
        )

    def initialize(
        self,
        grid_size=(5, 5),
        report_interval=5.0,
        grid_orientation="vertical",
        node_layout="rect",
        show_plots=False,
        cts_type="oriented_hex",
        run_duration=1.0,
        output_interval=1.0e99,
        plot_every_transition=False,
        **kwds
    ):

        # Remember the clock time, and calculate when we next want to report
        # progress.
        self.current_real_time = time.time()
        self.next_report = self.current_real_time + report_interval
        self.report_interval = report_interval

        # Interval for output
        self.output_interval = output_interval

        # Duration for run
        self.run_duration = run_duration

        # Create a grid
        self.create_grid_and_node_state_field(
            grid_size[0], grid_size[1], grid_orientation, node_layout, cts_type
        )

        # Create the node-state dictionary
        ns_dict = self.node_state_dictionary()

        # Initialize values of the node-state grid
        nsg = self.initialize_node_state_grid()

        # Create the transition list
        xn_list = self.transition_list()

        # Create the CA object
        if cts_type == "raster":
            from landlab.ca.raster_cts import RasterCTS

            self.ca = RasterCTS(self.grid, ns_dict, xn_list, nsg)
        elif cts_type == "oriented_raster":
            from landlab.ca.oriented_raster_cts import OrientedRasterCTS

            self.ca = OrientedRasterCTS(self.grid, ns_dict, xn_list, nsg)
        elif cts_type == "hex":
            from landlab.ca.hex_cts import HexCTS

            self.ca = HexCTS(self.grid, ns_dict, xn_list, nsg)
        else:
            from landlab.ca.oriented_hex_cts import OrientedHexCTS

            self.ca = OrientedHexCTS(self.grid, ns_dict, xn_list, nsg)

        # Initialize graphics
        self._show_plots = show_plots
        if show_plots:
            self.initialize_plotting(**kwds)

    def create_grid_and_node_state_field(
        self, num_rows, num_cols, grid_orientation, node_layout, cts_type
    ):
        """Create the grid and the field containing node states."""

        if cts_type == "raster" or cts_type == "oriented_raster":
            from landlab import RasterModelGrid

            self.grid = RasterModelGrid(shape=(num_rows, num_cols), xy_spacing=1.0)
        else:
            from landlab import HexModelGrid

            self.grid = HexModelGrid(
                num_rows,
                num_cols,
                xy_spacing=1.0,
                orientation=grid_orientation,
                node_layout=node_layout,
            )

        self.grid.add_zeros("node", "node_state", dtype=int)

    def node_state_dictionary(self):
        """Create and return a dictionary of all possible node (cell) states.

        This method creates a default set of states (just two); it is a
        template meant to be overridden.
        """
        ns_dict = {0: "on", 1: "off"}
        return ns_dict

    def transition_list(self):
        """Create and return a list of transition objects.

        This method creates a default set of transitions (just two); it is a
        template meant to be overridden.
        """
        xn_list = []
        xn_list.append(Transition((0, 1, 0), (1, 0, 0), 1.0))
        xn_list.append(Transition((1, 0, 0), (0, 1, 0), 1.0))
        return xn_list

    def write_output(self, grid, outfilename, iteration):
        """Write output to file (currently netCDF)."""
        filename = outfilename + str(iteration).zfill(4) + ".nc"
        save_grid(grid, filename)

    def initialize_node_state_grid(self):
        """Initialize values in the node-state grid.

        This method should be overridden. The default is random "on" and "off".
        """
        num_states = 2
        for i in range(self.grid.number_of_nodes):
            self.grid.at_node["node_state"][i] = random.randint(num_states)
        return self.grid.at_node["node_state"]

    def initialize_plotting(self, **kwds):
        """Create and configure CAPlotter object."""
        self.ca_plotter = CAPlotter(self.ca, **kwds)
        self.ca_plotter.update_plot()
        axis("off")

    def run_for(self, dt):

        self.ca.run(self.ca.current_time + dt, self.ca.node_state)


if __name__ == "__main__":
    ctsm = CTSModel(show_plots=True)
    ctsm.run_for(1.0)
    ctsm.ca_plotter.update_plot()



# -*- coding: utf-8 -*-
"""
test_fracture_grid.py

Unit test for fracture_grid component.

Created on Sun Oct 18 09:47:59 2015

@author: gtucker
"""

from numpy.testing import assert_array_equal

from landlab import RasterModelGrid
from landlab.components.fracture_grid import FractureGridGenerator


def test_frac_grid():
    nrows = 9
    ncols = 9

    grid = RasterModelGrid((nrows, ncols))
    fg = FractureGridGenerator(grid, frac_spacing=3, seed=1)
    fg.run_one_step()

    assert_array_equal(
        grid.at_node["fracture_at_node"].reshape((9, 9)),
        [
            [1, 0, 0, 0, 0, 0, 1, 1, 0],
            [0, 0, 0, 0, 1, 1, 0, 0, 0],
            [1, 1, 1, 1, 1, 1, 1, 0, 0],
            [1, 1, 1, 0, 0, 0, 0, 1, 1],
            [0, 0, 1, 1, 1, 1, 1, 1, 1],
            [1, 1, 1, 1, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 1, 1, 0, 0],
            [0, 0, 0, 0, 0, 1, 1, 1, 0],
            [0, 0, 0, 0, 0, 0, 0, 1, 1],
        ],
    )



#!/usr/bin/env python
"""Constants used with the netcdf module."""


_DIMENSION_NAMES = ["ni", "nj", "nk"]
_AXES_NAMES = ["x", "y", "z"]

_NP_TO_NC_TYPE = {
    "float32": "f4",
    "float64": "f8",
    "int8": "i1",
    "int16": "i2",
    "int32": "i4",
    "int64": "i8",
    "uint8": "u1",
    "uint16": "u2",
    "uint32": "u4",
    "uint64": "u8",
    "bool": "i1",
}


_AXIS_DIMENSION_NAMES = ["nk", "nj", "ni"]
_AXIS_COORDINATE_NAMES = ["z", "y", "x"]

_DIMENSION_NAMES = set(_AXIS_DIMENSION_NAMES + ["nt"])
_COORDINATE_NAMES = set(_AXIS_COORDINATE_NAMES + ["t"])


_GRID_MAPPING_VARIABLES = [
    "crs_wkt",
    "earth_radius",
    "false_easting",
    "false_northing",
    "grid_mapping_name",
    "grid_north_pole_latitude",
    "grid_north_pole_longitude",
    "inverse_flattening",
    "latitude_of_projection_origin",
    "longitude_of_central_meridian",
    "longitude_of_prime_meridian",
    "longitude_of_projection_origin",
    "north_pole_grid_longitude",
    "perspective_point_height",
    "scale_factor_at_central_meridian",
    "scale_factor_at_projection_origin",
    "semi_major_axis",
    "semi_minor_axis",
    "standard_parallel",
    "straight_vertical_longitude_from_pole",
]



"""
Unit tests for landlab.components.plant_competition_ca.plant_competition_ca
"""
import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal

(_SHAPE, _SPACING, _ORIGIN) = ((20, 20), (10e0, 10e0), (0.0, 0.0))
_ARGS = (_SHAPE, _SPACING, _ORIGIN)


def test_name(ca_veg):
    assert ca_veg.name == "Cellular Automata Plant Competition"


def test_input_var_names(ca_veg):
    assert sorted(ca_veg.input_var_names) == [
        "vegetation__cumulative_water_stress",
        "vegetation__plant_functional_type",
    ]


def test_output_var_names(ca_veg):
    assert sorted(ca_veg.output_var_names) == ["plant__age", "plant__live_index"]


def test_var_units(ca_veg):
    assert set(ca_veg.input_var_names) | set(ca_veg.output_var_names) == set(
        dict(ca_veg.units).keys()
    )

    assert ca_veg.var_units("vegetation__cumulative_water_stress") == "None"
    assert ca_veg.var_units("vegetation__plant_functional_type") == "None"
    assert ca_veg.var_units("plant__live_index") == "None"
    assert ca_veg.var_units("plant__age") == "Years"


def test_grid_shape(ca_veg):
    assert ca_veg.grid.number_of_node_rows == _SHAPE[0]
    assert ca_veg.grid.number_of_node_columns == _SHAPE[1]


def test_grid_x_extent(ca_veg):
    assert ca_veg.grid.extent[1] == (_SHAPE[1] - 1) * _SPACING[1]


def test_grid_y_extent(ca_veg):
    assert ca_veg.grid.extent[0] == (_SHAPE[0] - 1) * _SPACING[0]


def test_field_getters(ca_veg):
    for name in ca_veg.grid["node"]:
        field = ca_veg.grid["node"][name]
        assert isinstance(field, np.ndarray)
        assert field.shape == (
            ca_veg.grid.number_of_node_rows * ca_veg.grid.number_of_node_columns,
        )

    for name in ca_veg.grid["cell"]:
        field = ca_veg.grid["cell"][name]
        assert isinstance(field, np.ndarray)
        assert field.shape, (
            ca_veg.grid.number_of_cell_rows * ca_veg.grid.number_of_cell_columns,
        )

    with pytest.raises(KeyError):
        ca_veg.grid["not_a_var_name"]


def test_field_initialized_to_zero(ca_veg):
    for name in ca_veg.grid["node"]:
        field = ca_veg.grid["node"][name]
        assert_array_almost_equal(field, np.zeros(ca_veg.grid.number_of_nodes))



from __future__ import print_function

import numpy as np

from landlab.layers.eventlayers import (
    EventLayers,
    _deposit_or_erode,
    _get_surface_index,
)


class MaterialLayersMixIn(object):

    """MixIn that adds a MaterialLayers attribute to a ModelGrid."""

    @property
    def material_layers(self):
        """MaterialLayers for each cell."""
        try:
            self._material_layers
        except AttributeError:
            self._material_layers = MaterialLayers(self.number_of_cells)
        finally:
            return self._material_layers


class MaterialLayers(EventLayers):

    """Track MaterialLayers where each layer has some material in it.

    MaterialLayers are meant to represent a layered object in which each layer
    has some material in it. If erosion occurs, no new layer is created. These
    layers stand in contrast to the EventLayers for which each event is
    represented by a layer.

    MaterialLayers is likely a more memory efficent data structure than
    EventLayers as it does not record erosion as an array of zeros.

    Parameters
    ----------
    number_of_stacks : int
        Number of layer stacks to track.

    Attributes
    ----------
    allocated
    dz
    thickness
    number_of_layers
    number_of_stacks
    surface_index
    z

    Methods
    -------
    add
    get_surface_values

    Examples
    --------
    >>> from landlab.layers.materiallayers import MaterialLayers

    Create an empty layer stack with 5 stacks.

    >>> layers = MaterialLayers(5)
    >>> layers.number_of_stacks
    5
    >>> layers.number_of_layers
    0

    Add a layer with a uniform thickness.

    >>> layers.add(1.5)
    >>> layers.number_of_layers
    1
    >>> layers.dz
    array([[ 1.5,  1.5,  1.5,  1.5,  1.5]])

    MaterialLayers will combine layers if they have the same attributes.
    Adding a second layer with uneven thickness. Will increment the
    first layers thickness. This stands in contrast with EventLayers
    which will track each addition as a separate entry in the layers
    datastructure.

    >>> layers.add([1., 2., 3., 5., 0.])
    >>> layers.dz
    array([[ 2.5,  3.5,  4.5,  6.5,  1.5]])

    Adding a layer with negative thickness will remove
    material from the layers. Unlike EventLayers, it will not add a
    layer of zeros that represent an event with no deposition.

    >>> layers.add(-1)
    >>> layers.dz
    array([[ 1.5,  2.5,  3.5,  5.5,  0.5]])

    Get the index value of the layer within each stack
    at the topographic surface.

    >>> layers.surface_index
    array([0, 0, 0, 0, 0])

    See the example in the ``add`` method to learn how MaterialLayers
    behaves if material properties are also tracked.
    """

    def add(self, dz, **kwds):
        """Add a layer to the  MaterialLayers stacks.

        Parameters
        ----------
        dz : float or array_like
            Thickness to add to each stack.

        Examples
        --------
        >>> from landlab.layers.materiallayers import MaterialLayers

        Create an empty layer stack with 3 stacks.

        >>> layers = MaterialLayers(3)
        >>> layers.number_of_layers
        0

        To add a layer of uniform thickness to every stack.

        >>> layers.add(1.5)
        >>> layers.number_of_layers
        1
        >>> layers.dz
        array([[ 1.5,  1.5,  1.5]])

        Add a second layer with uneven thickness.

        >>> layers.add([1., 2., .5])
        >>> layers.dz
        array([[ 2.5,  3.5,  2. ]])

        Because the attributes of this layer and the previous layer
        are the same (e.g. they don't exist), MaterialLayer will combined
        them. This is the primary difference between MaterialLayers and
        EventLayers.

        Adding a layer with negative thickness will remove material from
        the top of the stack.

        >>> layers.add(-1)
        >>> layers.dz
        array([[ 1.5,  2.5,  1. ]])
        >>> layers.number_of_layers
        1

        Use keywords to track properties of each layer. For instance,
        here we create a new stack and add a layer with a particular
        *type* and a particular *size*. You can access the layer properties as
        if the object were a dictionary.

        >>> layers = MaterialLayers(3)
        >>> layers.add(1., type=3., size='sand')
        >>> layers.dz
        array([[ 1.,  1.,  1.]])
        >>> layers['type']
        array([[ 3.,  3.,  3.]])

        As you can see, there is no rule that says you can't use a string as
        the value of an attribute.

        Adding a layer with the same attributes as the entire surface of the
        MaterialLayers will result in the layers being combined.

        >>> layers.add(1., type=3., size='sand')
        >>> layers.add([2, -1, 0], type=3., size='sand')
        >>> layers.dz
        array([[ 4.,  1.,  2.]])

        Adding material with different attributes results in the creation of
        a new layer.

        >>> layers.add(2., type=6., size='sand')
        >>> layers.dz
        array([[ 4.,  1.,  2.],
               [ 2.,  2.,  2.]])
        >>> layers['type']
        array([[ 3.,  3.,  3.],
               [ 6.,  6.,  6.]])
        >>> np.all(layers['size'] == [['sand', 'sand', 'sand'],
        ...                           ['sand', 'sand', 'sand']])
        True

        array([['sand', 'sand', 'sand'],
               ['sand', 'sand', 'sand']],
              dtype='<U4')

        Attributes for each layer will exist even if part the the layer is
        associated with erosion.

        >>> layers.add([-2, -1, 1], type=8., size='gravel')
        >>> layers.dz
        array([[ 4.,  1.,  2.],
               [ 0.,  1.,  2.],
               [ 0.,  0.,  1.]])
        >>> layers['type']
        array([[ 3.,  3.,  3.],
               [ 6.,  6.,  6.],
               [ 8.,  8.,  8.]])

        To get the values at the surface of the layer stack:

        >>> layers.get_surface_values('type')
        array([ 3.,  6.,  8.])

        Removing enough material such that an entire layer's
        thickness is no longer present, results in that layer
        no longer being tracked. This is another difference
        between MaterialLayers and EventLayers.

        >>> layers.add([ 0., 0., -1. ])
        >>> layers.dz
        array([[ 4.,  1.,  2.],
               [ 0.,  1.,  2.]])
        >>> layers['type']
        array([[ 3.,  3.,  3.],
               [ 6.,  6.,  6.]])
        >>> np.all(layers['size'] == [['sand', 'sand', 'sand'],
        ...                           ['sand', 'sand', 'sand']])
        True
        >>> layers.number_of_layers
        2

        If attributes (like age and size in this example) are tracked, a layer
        will be combined with the surface layer only if all attributes are the
        same across the entire layer. Right now, the surface values vary.

        >>> layers.get_surface_values('type')
        array([ 3.,  6.,  6.])
        >>> np.all(layers.get_surface_values('size') == ['sand', 'sand', 'sand'])
        True

        Since the surface has different types, adding material will create a
        new layer.

        >>> layers.add(3., type=6., size='sand')
        >>> layers.dz
        array([[ 4.,  1.,  2.],
               [ 0.,  1.,  2.],
               [ 3.,  3.,  3.]])
        >>> layers['type']
        array([[ 3.,  3.,  3.],
               [ 6.,  6.,  6.],
               [ 6.,  6.,  6.]])
        >>> layers.number_of_layers
        3

        But now, the entire surface has the qualities of type = 6. and size =
        'sand', so layers will be combined. This even works if the thickness of
        the new layer includes both erosion and deposition.

        >>> layers.add([ -3.5, 0., 2. ], type=6., size='sand')
        >>> layers.dz
        array([[ 3.5,  1. ,  2. ],
               [ 0. ,  1. ,  2. ],
               [ 0. ,  3. ,  5. ]])
        >>> layers['type']
        array([[ 3.,  3.,  3.],
               [ 6.,  6.,  6.],
               [ 6.,  6.,  6.]])
        >>> layers.number_of_layers
        3
        """
        dz = np.asarray(dz)

        if self.number_of_layers == 0:
            self._setup_layers(**kwds)

        compatible = self.number_of_layers > 0 and self.is_compatible(dz, **kwds)

        if not compatible:
            self._add_empty_layer()

        _deposit_or_erode(self._attrs["_dz"], self.number_of_layers, dz)
        _get_surface_index(
            self._attrs["_dz"], self.number_of_layers, self._surface_index
        )

        self._remove_empty_layers()

        if not compatible:
            for name in kwds:
                self[name][-1] = kwds[name]

    def _remove_empty_layers(self):
        number_of_filled_layers = self.surface_index.max() + 1
        if number_of_filled_layers < self.number_of_layers:
            self._number_of_layers = number_of_filled_layers

    def is_compatible(self, dz, **kwds):
        """Check if a new layer is compatible with the existing top layer.

        Parameters
        ----------
        dz : float or array_like
            Thickness to add to each stack.

        Returns
        -------
        bool
            ``True`` if the new layer is compatible, otherwise ``False``.
        """
        where_deposition = np.where(dz > 0.0)[0]
        if len(where_deposition) > 0:
            for name in kwds:
                try:
                    is_compatible = self[name][self.surface_index] == kwds[name]
                except KeyError:
                    msg = "MaterialLayers: {0} is not being tracked. Error in adding.".format(
                        name
                    )
                    raise ValueError(msg)

                if not np.all(is_compatible[where_deposition]):
                    return False
        return True



#! /usr/bin/env python
"""Create landlab model grids."""
from __future__ import absolute_import

from warnings import warn

from ..core import load_params, model_parameter_dictionary as mpd
from ..io import read_esri_ascii
from ..io.netcdf import read_netcdf
from ..values import constant, plane, random, sine
from .hex import HexModelGrid, from_dict as hex_from_dict
from .network import NetworkModelGrid
from .radial import RadialModelGrid
from .raster import RasterModelGrid, from_dict as raster_from_dict
from .voronoi import VoronoiDelaunayGrid

_MODEL_GRIDS = {
    "RasterModelGrid": RasterModelGrid,
    "HexModelGrid": HexModelGrid,
    "VoronoiDelaunayGrid": VoronoiDelaunayGrid,
    "NetworkModelGrid": NetworkModelGrid,
    "RadialModelGrid": RadialModelGrid,
}

_SYNTHETIC_FIELD_CONSTRUCTORS = {
    "plane": plane,
    "random": random,
    "sine": sine,
    "constant": constant,
}


class Error(Exception):

    """Base class for exceptions from this module."""

    pass


class BadGridTypeError(Error):

    """Raise this error for a bad grid type."""

    def __init__(self, grid_type):
        self._type = str(grid_type)  # TODO: not tested.

    def __str__(self):
        return self._type  # TODO: not tested.


_GRID_READERS = {"raster": raster_from_dict, "hex": hex_from_dict}


def create_and_initialize_grid(input_source):
    """Create and initialize a grid from a file.

    Creates, initializes, and returns a new grid object using parameters
    specified in *input_source*. *input_source* is either a
    ModelParameterDictionary instance (or, really, just dict-like) or a
    named input file.

    Parameters
    ----------
    input_source : str or dict
        Input file or ``dict`` of parameter values.

    Raises
    ------
    KeyError
        If missing a key from the input file.

    Examples
    --------
    >>> from six import StringIO
    >>> import pytest
    >>> test_file = StringIO('''
    ... GRID_TYPE:
    ... raster
    ... NUM_ROWS:
    ... 4
    ... NUM_COLS:
    ... 5
    ... GRID_SPACING:
    ... 2.5
    ... ''')
    >>> from landlab import create_and_initialize_grid
    >>> with pytest.deprecated_call():
    ...    grid = create_and_initialize_grid(test_file)
    >>> grid.number_of_nodes
    20
    """
    msg = (
        "create_and_initialize_grid is deprecated and will be removed "
        "in landlab 2.0. Use create_grid instead."
    )
    warn(msg, DeprecationWarning)
    if isinstance(input_source, dict):
        param_dict = input_source  # TODO: not tested.
    else:
        param_dict = mpd.ModelParameterDictionary(from_file=input_source)

    grid_type = param_dict["GRID_TYPE"]

    grid_type.strip().lower()

    # Read parameters appropriate to that type, create it, and initialize it
    try:
        grid_reader = _GRID_READERS[grid_type]
    except KeyError:  # TODO: not tested.
        raise BadGridTypeError(grid_type)  # TODO: not tested.

    # Return the created and initialized grid
    return grid_reader(param_dict)


def grid_from_dict(grid_type, params):
    """Create a grid from a dictionary of parameters."""
    try:
        cls = _MODEL_GRIDS[grid_type]
    except KeyError:
        raise ValueError("unknown grid type ({0})".format(grid_type))
    args, kwargs = _parse_args_kwargs(params)
    return cls(*args, **kwargs)


def grids_from_file(file_like, section=None):
    """Create grids from a file."""
    params = load_params(file_like)

    if section:
        try:
            grids = params[section]
        except KeyError:  # TODO: not tested.
            raise ValueError(
                "missing required section ({0})".format(section)
            )  # TODO: not tested.
    else:  # TODO: not tested.
        grids = params  # TODO: not tested.

    new_grids = []
    for grid_type, grid_desc in as_list_of_tuples(grids):
        new_grids.append(grid_from_dict(grid_type, grid_desc))

    return new_grids


def add_fields_from_dict(grid, fields):
    """Add fields to a grid from a dictionary."""
    fields = dict(fields)

    unknown_locations = set(fields) - set(grid.VALID_LOCATIONS)
    if unknown_locations:
        raise ValueError(
            "unknown field locations ({0})".format(", ".join(unknown_locations))
        )

    for location, fields_at_location in fields.items():
        for name, function in fields_at_location.items():
            add_field_from_function(grid, name, function, at=location)

    return grid


def add_field_from_function(grid, name, functions, at="node"):
    """Add a field to a grid as functions.

    Parameters
    ----------
    grid : ModelGrid
        A landlab grid to add fields to.
    name : str
        Name of the new field.
    functions : *(func_name, func_args)* or iterable of *(func_name, func_args)*
        The functions to apply to the field. Functions are applied in the order
        the appear in the list.
    at : str
        The grid element to which the field will be added.

    Returns
    -------
    ModelGrid
        The grid with the new field.
    """
    valid_functions = set(_SYNTHETIC_FIELD_CONSTRUCTORS) | set(
        ["read_esri_ascii", "read_netcdf"]
    )

    for func_name, func_args in as_list_of_tuples(functions):
        if func_name not in valid_functions:
            raise ValueError("function not understood ({0})".format(func_name))

        args, kwargs = _parse_args_kwargs(func_args)

        if func_name in _SYNTHETIC_FIELD_CONSTRUCTORS:
            # if any args, raise an error, there shouldn't be any.
            synth_function = _SYNTHETIC_FIELD_CONSTRUCTORS[func_name]
            synth_function(grid, name, at=at, **kwargs)
        elif func_name == "read_esri_ascii":
            read_esri_ascii(*args, grid=grid, name=name, **kwargs)
        elif func_name == "read_netcdf":
            read_netcdf(*args, grid=grid, name=name, **kwargs)

    return grid


def add_boundary_conditions(grid, boundary_conditions=()):
    for bc_name, bc_args in as_list_of_tuples(boundary_conditions):
        args, kwargs = _parse_args_kwargs(bc_args)
        try:
            func = getattr(grid, bc_name)
        except AttributeError:
            raise ValueError(
                "create_grid: No function {func} exists for grid types {grid}."
                "If you think this type of grid should have such a "
                "function. Please create a GitHub Issue to discuss "
                "contributing it to the Landlab codebase.".format(
                    func=bc_name, grid=grid.__class__.__name__
                )
            )
        else:
            func(*args, **kwargs)


def as_list_of_tuples(items):
    """Convert a collection of key/values to a list of tuples.

    Examples
    --------
    >>> from collections import OrderedDict
    >>> from landlab.grid.create import as_list_of_tuples
    >>> as_list_of_tuples({"eric": "idle"})
    [('eric', 'idle')]
    >>> as_list_of_tuples([("john", "cleese"), {"eric": "idle"}])
    [('john', 'cleese'), ('eric', 'idle')]
    >>> as_list_of_tuples(
    ...     [("john", "cleese"), OrderedDict([("eric", "idle"), ("terry", "gilliam")])]
    ... )
    [('john', 'cleese'), ('eric', 'idle'), ('terry', 'gilliam')]
    """
    try:
        items = list(items.items())
    except AttributeError:
        items = list(items)

    if len(items) == 2 and isinstance(items[0], str):
        items = [items]

    tuples = []
    for item in items:
        try:
            tuples.extend(list(item.items()))
        except AttributeError:
            tuples.append(tuple(item))

    return tuples


def create_grid(file_like, section=None):
    """Create grid, initialize fields, and set boundary conditions.

    **create_grid** expects a dictionary with three keys: "grid", "fields", and
    "boundary_conditions".

    **Dictionary Section "grid"**

    The value associated with the "grid" key should itself be a dictionary
    containing the name of a Landlab model grid type as its only key. The
    following grid types are valid:

        -  :py:class:`~landlab.grid.raster.RasterModelGrid`
        -  :py:class:`~landlab.grid.voronoi.VoronoiDelaunayGrid`
        -  :py:class:`~landlab.grid.hex.HexModelGrid`
        -  :py:class:`~landlab.grid.radial.RadialModelGrid`
        -  :py:class:`~landlab.grid.network.NetworkModelGrid`

    The value associated with the grid name key is a list containing the
    arguments. If any keyword arguments are passed, they should be passed as
    the last element of the list. For example the following code block is a
    yaml file indicating a RasterModelGrid with shape (4, 5) and xy-spacing of
    (3, 4).

    .. code-block:: yaml

        grid:
          RasterModelGrid:
            - [4, 5]
            - xy_spacing: [3, 4]

    These arguments and keyword arguments will be passed to the ``__init__``
    constructor of the specified model grid. Refer to the documentation for
    each grid to determine its requirements.

    **Dictionary Section "fields"**

    Fields can be created by reading from files or by creating synthetic
    values.

    The value associated with the "fields" key is a nested set of dictionaries
    indicating where the fields are created, what the field names are, and how
    to create the fields. As part of a grid's description, the value
    associated with the "fields" key must be a dictionary with keys indicating
    at which grid elements fields should be created (e.g. to create fields at
    node, use "node").

    The value associated with each "xxx" (i.e. "node", "link", "patch", etc.)
    value is itself a dictionary
    indicating the name of the field and how it should be created. A field can
    either be created by reading from a file or creating synthetic values. The
    :py:func:`~landlab.io.netcdf.read.read_netcdf` and
    :py:func:`~landlab.io.esri_ascii.read_esri_ascii` functions, and the
    :py:mod:`synthetic fields <landlab.values.synthetic>`
    package are currently supported methods to create fields. These may be
    chained together (as is shown in the Example section below). If these
    functions do not meet your needs, we welcome contributions that extend the
    capabilities of this function.

    The following example would use the
    :py:func:`~landlab.values.synthetic.plane` function from the synthetic
    values package to create a *node* value for the field
    *topographic__elevation*. The plane function adds values to a Landlab model
    grid field that lie on a plane specified by a point and a normal vector. In
    the below example the plane goes through the point (1.0, 1.0, 1.0) and has
    a normal of (-2.0, -1.0, 1.0).

    .. code-block:: yaml

        grid:
          RasterModelGrid:
            - [4, 5]
            - xy_spacing: [3, 4]
            - fields:
                node:
                  topographic__elevation:
                    plane:
                      - point: [1, 1, 1]
                        normal: [-2, -1, 1]

    **Dictionary Section "boundary_conditions"**

    The final portion of the input dictionary calls bound functions of the
    model grid to set boundary conditions. Any valid bound function can be
    called. The specified functions are provided in a list, and called in
    order. If required, multiple functions may be called.

    Each entry to the list is a dictionary with a single key, the name of the
    bound function. The value associated with that key is a list of arguments
    and keyword arguments, similar in structure to those described above.
    As with the "fields" section, the "boundary_conditions" section must be
    described under its associated grid description.

    For example, the following sets closed boundaries at all sides of the grid.

    .. code-block:: yaml

        grid:
          RasterModelGrid:
            - [4, 5]
            - xy_spacing: [3, 4]
            - boundary_conditions:
              - set_closed_boundaries_at_grid_edges:
                - True
                - True
                - True
                - True

    Parameters
    ----------
    file_like : file_like or str
        Dictionary, contents of a dictionary as a string, a file-like object,
        or the path to a file containing a YAML dictionary.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import create_grid
    >>> np.random.seed(42)
    >>> p = {
    ...     "grid": {
    ...         "RasterModelGrid": [
    ...             (4, 5),
    ...             {"xy_spacing": (3, 4)},
    ...             {
    ...                 "fields": {
    ...                     "node": {
    ...                         "spam": {
    ...                             "plane": [{"point": (1, 1, 1), "normal": (-2, -1, 1)}],
    ...                             "random": [
    ...                                 {"distribution": "uniform", "low": 1, "high": 4}
    ...                             ],
    ...                         }
    ...                     },
    ...                     "link": {
    ...                         "eggs": {"constant": [{"where": "ACTIVE_LINK", "value": 12}]}
    ...                     },
    ...                 }
    ...             },
    ...             {
    ...                 "boundary_conditions": [
    ...                     {"set_closed_boundaries_at_grid_edges": [True, True, True, True]}
    ...                 ]
    ...             },
    ...         ]
    ...     }
    ... }
    >>> mg = create_grid(p, section="grid")
    >>> mg.number_of_nodes
    20
    >>> "spam" in mg.at_node
    True
    >>> "eggs" in mg.at_link
    True
    >>> mg.x_of_node
    array([  0.,   3.,   6.,   9.,  12.,
             0.,   3.,   6.,   9.,  12.,
             0.,   3.,   6.,   9.,  12.,
             0.,   3.,   6.,   9.,  12.])
    >>> mg.status_at_node
    array([4, 4, 4, 4, 4,
           4, 0, 0, 0, 4,
           4, 0, 0, 0, 4,
           4, 4, 4, 4, 4], dtype=uint8)
    >>> np.round(mg.at_node['spam'].reshape(mg.shape), decimals=2)
    array([[  0.12,   7.85,  13.2 ,  18.8 ,  23.47],
           [  3.47,   9.17,  17.6 ,  22.8 ,  29.12],
           [  7.06,  15.91,  21.5 ,  25.64,  31.55],
           [ 11.55,  17.91,  24.57,  30.3 ,  35.87]])
    """
    if isinstance(file_like, dict):
        params = file_like
    else:
        params = load_params(file_like)

    if section:
        grids = params[section]
    else:
        grids = params

    new_grids = []
    for grid_type, grid_desc in as_list_of_tuples(grids):
        grid_desc = norm_grid_description(grid_desc)

        fields = grid_desc.pop("fields", {})
        boundary_conditions = grid_desc.pop("boundary_conditions", {})

        grid = grid_from_dict(grid_type, grid_desc)
        add_fields_from_dict(grid, fields)
        add_boundary_conditions(grid, boundary_conditions)

        new_grids.append(grid)

    if len(new_grids) == 1:
        return new_grids[0]
    else:
        return new_grids


def norm_grid_description(grid_desc):
    """Normalize a grid description into a canonical form.

    Examples
    --------
    >>> from landlab.grid.create import norm_grid_description

    >>> grid_desc = [
    ...     (3, 4), {"xy_spacing": 4.0, "xy_of_lower_left": (1.0, 2.0)}
    ... ]
    >>> normed_items = list(norm_grid_description(grid_desc).items())
    >>> normed_items.sort()
    >>> normed_items
    [('args', [(3, 4)]), ('xy_of_lower_left', (1.0, 2.0)), ('xy_spacing', 4.0)]
    """
    if not isinstance(grid_desc, dict):
        args, kwds = [], {}
        for arg in grid_desc:
            if isinstance(arg, dict) and {"fields", "boundary_conditions"} & set(
                arg.keys()
            ):
                kwds.update(arg)
            else:
                args.append(arg)
        if isinstance(args[-1], dict):
            kwds.update(args.pop())
        kwds.update({"args": args})
        return kwds
    return grid_desc


def _parse_args_kwargs(list_of_args_kwargs):
    if isinstance(list_of_args_kwargs, dict):
        args, kwargs = list_of_args_kwargs.pop("args", ()), list_of_args_kwargs
        if not isinstance(args, (tuple, list)):
            args = (args,)
    else:
        args, kwargs = [], {}
        for arg in list(list_of_args_kwargs):
            if isinstance(arg, dict) and {"fields", "boundary_conditions"} & set(
                arg.keys()
            ):
                kwargs.update(arg)  # TODO: not tested.
            else:
                args.append(arg)
        if isinstance(args[-1], dict):
            kwargs.update(args.pop())

    return tuple(args), kwargs



import numpy as np
import pytest
from numpy.testing import assert_array_equal
from six import StringIO

from landlab import (
    CLOSED_BOUNDARY,
    HexModelGrid,
    NetworkModelGrid,
    RadialModelGrid,
    RasterModelGrid,
    VoronoiDelaunayGrid,
)
from landlab.grid.hex import from_dict as hex_from_dict
from landlab.grid.raster import from_dict as raster_from_dict


def test_raster_old_from_dict_deprecated():
    params = {"NUM_COLS": 10, "NUM_ROWS": 5, "GRID_SPACING": 4}
    with pytest.deprecated_call():
        mg = raster_from_dict(params)
    assert mg.shape == (5, 10)
    assert mg.dx == 4
    assert mg.dy == 4


def test_hex_old_from_dict_deprecated():
    params = {"NUM_COLS": 10, "NUM_ROWS": 5, "GRID_SPACING": 4}
    with pytest.deprecated_call():
        mg = hex_from_dict(params)
    assert mg.number_of_nodes == 54


def test_raster_from_file():
    file_strn = (
        "shape:\n"
        "    - 10\n"
        "    - 20\n"
        "xy_spacing:\n"
        "    - 25\n"
        "    - 45\n"
        "bc:\n"
        "    right: 'closed'\n"
        "    top: 'closed'\n"
        "    left: 'closed'\n"
        "    bottom: 'closed'\n"
        "xy_of_reference:\n"
        "    - 12345\n"
        "    - 678910\n"
        "xy_of_lower_left:\n"
        "    - 35\n"
        "    - 55\n"
        "axis_name:\n"
        "    - 'spam'\n"
        "    - 'eggs'\n"
        "axis_units:\n"
        "    - 'smoot'\n"
        "    - 'parsec'"
    )
    file_like = StringIO(file_strn)
    mg = RasterModelGrid.from_file(file_like)

    # assert things.
    assert mg.shape == (10, 20)
    assert mg.dx == 25
    assert mg.dy == 45
    assert (mg.x_of_node.min(), mg.y_of_node.min()) == (35, 55)
    assert np.all(mg.status_at_node[mg.boundary_nodes] == CLOSED_BOUNDARY)
    assert mg.axis_units == ("smoot", "parsec")
    assert mg.axis_name == ("spam", "eggs")
    assert mg.xy_of_reference == (12345, 678910)


def test_raster_from_dict():
    params = {
        "shape": (10, 20),
        "xy_spacing": (25, 45),
        "bc": {
            "right": "closed",
            "top": "closed",
            "left": "closed",
            "bottom": "closed",
        },
        "xy_of_lower_left": (35, 55),
        "axis_name": ("spam", "eggs"),
        "axis_units": ("smoot", "parsec"),
        "xy_of_reference": (12345, 678910),
    }

    mg = RasterModelGrid.from_dict(params)

    # assert things.
    assert mg.shape == (10, 20)
    assert mg.dx == 25
    assert mg.dy == 45
    assert (mg.x_of_node.min(), mg.y_of_node.min()) == (35, 55)
    assert np.all(mg.status_at_node[mg.boundary_nodes] == CLOSED_BOUNDARY)
    assert mg.axis_units == ("smoot", "parsec")
    assert mg.axis_name == ("spam", "eggs")
    assert mg.xy_of_reference == (12345, 678910)


def test_hex_from_dict():
    params = {
        "base_num_rows": 5,
        "base_num_cols": 4,
        "dx": 2.0,
        "xy_of_lower_left": (35, 55),
        "axis_name": ("spam", "eggs"),
        "axis_units": ("smoot", "parsec"),
        "xy_of_reference": (12345, 678910),
    }

    mg = HexModelGrid.from_dict(params)

    # assert things.
    true_x_node = np.array(
        [
            37.0,
            39.0,
            41.0,
            43.0,
            36.0,
            38.0,
            40.0,
            42.0,
            44.0,
            35.0,
            37.0,
            39.0,
            41.0,
            43.0,
            45.0,
            36.0,
            38.0,
            40.0,
            42.0,
            44.0,
            37.0,
            39.0,
            41.0,
            43.0,
        ]
    )
    assert_array_equal(true_x_node, mg.x_of_node)
    assert (mg.x_of_node.min(), mg.y_of_node.min()) == (35, 55)
    assert mg.axis_units == ("smoot", "parsec")
    assert mg.axis_name == ("spam", "eggs")
    assert mg.xy_of_reference == (12345, 678910)


def test_radial_from_dict():
    params = {
        "num_shells": 5,
        "dr": 2.0,
        "xy_of_center": (35, 55),
        "axis_name": ("spam", "eggs"),
        "axis_units": ("smoot", "parsec"),
        "xy_of_reference": (12345, 678910),
    }

    mg = RadialModelGrid.from_dict(params)

    # assert things.
    assert mg.number_of_nodes == 95
    assert mg.xy_of_center == (35, 55)
    assert [35, 55] in mg.xy_of_node
    assert mg.axis_units == ("smoot", "parsec")
    assert mg.axis_name == ("spam", "eggs")
    assert mg.xy_of_reference == (12345, 678910)


def test_network_from_dict():
    params = {
        "yx_of_node": [(0, 1, 2, 2), (0, 0, -1, 1)],
        "links": ((1, 0), (2, 1), (3, 1)),
        "axis_name": ("spam", "eggs"),
        "axis_units": ("smoot", "parsec"),
        "xy_of_reference": (12345, 678910),
    }
    mg = NetworkModelGrid.from_dict(params)
    assert_array_equal(mg.x_of_node, np.array([0.0, 0.0, -1.0, 1.0]))
    assert_array_equal(mg.y_of_node, np.array([0.0, 1.0, 2.0, 2.0]))
    assert_array_equal(mg.nodes_at_link, np.array([[0, 1], [2, 1], [1, 3]]))
    assert mg.axis_units == ("smoot", "parsec")
    assert mg.axis_name == ("spam", "eggs")
    assert mg.xy_of_reference == (12345, 678910)


def test_network_from_file():
    file_strn = (
        "yx_of_node:\n"
        "    - [0, 1, 2, 2]\n"
        "    - [0, 0, -1, 1]\n"
        "links:\n"
        "    - [1, 0]\n"
        "    - [2, 1]\n"
        "    - [3, 1]\n"
        "xy_of_reference:\n"
        "    - 12345\n"
        "    - 678910\n"
        "axis_name:\n"
        "    - 'spam'\n"
        "    - 'eggs'\n"
        "axis_units:\n"
        "    - 'smoot'\n"
        "    - 'parsec'"
    )
    file_like = StringIO(file_strn)
    mg = NetworkModelGrid.from_file(file_like)

    assert_array_equal(mg.x_of_node, np.array([0.0, 0.0, -1.0, 1.0]))
    assert_array_equal(mg.y_of_node, np.array([0.0, 1.0, 2.0, 2.0]))
    assert_array_equal(mg.nodes_at_link, np.array([[0, 1], [2, 1], [1, 3]]))
    assert mg.axis_units == ("smoot", "parsec")
    assert mg.axis_name == ("spam", "eggs")
    assert mg.xy_of_reference == (12345, 678910)


def test_voronoi_from_dict():
    x = [0, 0.1, 0.2, 0.3, 1, 1.1, 1.2, 1.3, 2, 2.1, 2.2, 2.3]
    y = [0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3]
    params = {
        "x": x,
        "y": y,
        "axis_name": ("spam", "eggs"),
        "axis_units": ("smoot", "parsec"),
        "xy_of_reference": (12345, 678910),
    }

    mg = VoronoiDelaunayGrid.from_dict(params)
    assert mg.axis_units == ("smoot", "parsec")
    assert mg.axis_name == ("spam", "eggs")
    assert mg.xy_of_reference == (12345, 678910)

    true_x = np.array([0.0, 1.0, 2.0, 0.1, 1.1, 2.1, 0.2, 1.2, 2.2, 0.3, 1.3, 2.3])
    true_y = np.array([0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 2.0, 2.0, 2.0, 3.0, 3.0, 3.0])
    true_nodes_at_node = np.array(
        [
            [1, 3, -1, -1, -1, -1],
            [2, 4, 3, 0, -1, -1],
            [5, 4, 1, -1, -1, -1],
            [4, 6, 0, 1, -1, -1],
            [5, 7, 6, 3, 1, 2],
            [8, 7, 4, 2, -1, -1],
            [7, 9, 3, 4, -1, -1],
            [8, 10, 9, 6, 4, 5],
            [11, 10, 7, 5, -1, -1],
            [10, 6, 7, -1, -1, -1],
            [11, 9, 7, 8, -1, -1],
            [10, 8, -1, -1, -1, -1],
        ]
    )
    assert_array_equal(mg.node_x, true_x)
    assert_array_equal(mg.node_y, true_y)
    assert_array_equal(mg.adjacent_nodes_at_node, true_nodes_at_node)



import numpy as np
import pytest
from numpy.testing import assert_array_equal

from landlab import FieldError, RasterModelGrid
from landlab.components import DrainageDensity, FastscapeEroder, FlowAccumulator


def test_route_to_multiple_error_raised():
    mg = RasterModelGrid((10, 10))
    z = mg.add_zeros("node", "topographic__elevation")
    z += mg.x_of_node + mg.y_of_node
    fa = FlowAccumulator(mg, flow_director="MFD")
    fa.run_one_step()

    channel__mask = mg.zeros(at="node")

    with pytest.raises(NotImplementedError):
        DrainageDensity(mg, channel__mask=channel__mask)


def test_mask_is_stable():
    mg = RasterModelGrid((10, 10))
    mg.add_zeros("node", "topographic__elevation")
    np.random.seed(3542)
    noise = np.random.rand(mg.size("node"))
    mg.at_node["topographic__elevation"] += noise
    fr = FlowAccumulator(mg, flow_director="D8")
    fsc = FastscapeEroder(mg, K_sp=0.01, m_sp=0.5, n_sp=1)
    for x in range(2):
        fr.run_one_step()
        fsc.run_one_step(dt=10.0)
        mg.at_node["topographic__elevation"][mg.core_nodes] += 0.01

    mask = np.zeros(len(mg.at_node["topographic__elevation"]), dtype=np.uint8)
    mask[np.where(mg.at_node["drainage_area"] > 0)] = 1

    mask0 = mask.copy()

    dd = DrainageDensity(mg, channel__mask=mask)
    mask1 = mask.copy()

    dd.calc_drainage_density()
    mask2 = mask.copy()

    assert_array_equal(mask0, mask1)
    assert_array_equal(mask0[mg.core_nodes], mask2[mg.core_nodes])


def test_float_mask():
    mg = RasterModelGrid((10, 10))
    mg.add_zeros("node", "topographic__elevation")
    noise = np.random.rand(mg.size("node"))
    mg.at_node["topographic__elevation"] += noise
    fr = FlowAccumulator(mg, flow_director="D8")
    fr.run_one_step()
    mask = np.zeros(len(mg.at_node["topographic__elevation"]), dtype=float)
    mask[np.where(mg.at_node["drainage_area"] > 5)] = 1
    with pytest.raises(ValueError):
        DrainageDensity(mg, channel__mask=mask)


def test_bool_mask():
    mg = RasterModelGrid((10, 10))
    mg.add_zeros("node", "topographic__elevation")
    noise = np.random.rand(mg.size("node"))
    mg.at_node["topographic__elevation"] += noise
    fr = FlowAccumulator(mg, flow_director="D8")
    fr.run_one_step()
    mask = np.zeros(len(mg.at_node["topographic__elevation"]), dtype=bool)
    mask[np.where(mg.at_node["drainage_area"] > 5)] = 1
    with pytest.raises(ValueError):
        DrainageDensity(mg, channel__mask=mask)


def test_missing_fields():
    mg = RasterModelGrid((10, 10))
    mg.add_zeros("node", "topographic__elevation")
    with pytest.raises(FieldError):
        DrainageDensity(
            mg,
            area_coefficient=1,
            slope_coefficient=1,
            area_exponent=1,
            slope_exponent=1,
            channelization_threshold=1,
        )


def test_updating_with_array_provided():
    mg = RasterModelGrid((10, 10))
    mg.add_zeros("node", "topographic__elevation")
    noise = np.random.rand(mg.size("node"))
    mg.at_node["topographic__elevation"] += noise
    fr = FlowAccumulator(mg, flow_director="D8")
    fr.run_one_step()
    mask = np.zeros(len(mg.at_node["topographic__elevation"]), dtype=np.uint8)
    mask[np.where(mg.at_node["drainage_area"] > 5)] = 1
    dd = DrainageDensity(mg, channel__mask=mask)
    with pytest.raises(NotImplementedError):
        dd._update_channel_mask()


def test_mask_field_exists():
    mg = RasterModelGrid((10, 10))
    mg.add_zeros("node", "topographic__elevation")
    mg.add_zeros("node", "channel__mask")
    noise = np.random.rand(mg.size("node"))
    mg.at_node["topographic__elevation"] += noise
    fr = FlowAccumulator(mg, flow_director="D8")
    fr.run_one_step()
    mask = np.zeros(len(mg.at_node["topographic__elevation"]), dtype=np.uint8)
    mask[np.where(mg.at_node["drainage_area"] > 5)] = 1
    with pytest.warns(UserWarning):
        DrainageDensity(mg, channel__mask=mask)


def test_bad_mask_size():
    mg = RasterModelGrid((10, 10))
    mg.add_zeros("node", "topographic__elevation")
    noise = np.random.rand(mg.size("node"))
    mg.at_node["topographic__elevation"] += noise
    fr = FlowAccumulator(mg, flow_director="D8")
    fr.run_one_step()
    mask = np.zeros(20)
    mask[np.where(mg.at_node["drainage_area"][:20] > 5)] = 1

    with pytest.raises(ValueError):
        DrainageDensity(mg, channel__mask=mask)


def test_providing_array_and_kwargs():
    mg = RasterModelGrid((10, 10))
    mg.add_zeros("node", "topographic__elevation")
    noise = np.random.rand(mg.size("node"))
    mg.at_node["topographic__elevation"] += noise
    fr = FlowAccumulator(mg, flow_director="D8")
    fr.run_one_step()
    mask = np.zeros(len(mg.at_node["topographic__elevation"]), dtype=np.uint8)
    mask[np.where(mg.at_node["drainage_area"] > 5)] = 1
    with pytest.warns(UserWarning):
        DrainageDensity(mg, channel__mask=mask, area_coefficient=1)
    with pytest.warns(UserWarning):
        DrainageDensity(mg, channel__mask=mask, slope_coefficient=1)
    with pytest.warns(UserWarning):
        DrainageDensity(mg, channel__mask=mask, area_exponent=1)
    with pytest.warns(UserWarning):
        DrainageDensity(mg, channel__mask=mask, slope_exponent=1)
    with pytest.warns(UserWarning):
        DrainageDensity(mg, channel__mask=mask, channelization_threshold=1)


def test_missing_channel_threshold():
    mg = RasterModelGrid((10, 10))
    mg.add_zeros("node", "topographic__elevation")
    noise = np.random.rand(mg.size("node"))
    mg.at_node["topographic__elevation"] += noise
    fr = FlowAccumulator(mg, flow_director="D8")
    fr.run_one_step()
    with pytest.raises(ValueError):
        DrainageDensity(
            mg,
            area_coefficient=1,
            slope_coefficient=1,
            area_exponent=1,
            slope_exponent=1,
        )


def test_missing_slope_exponent():
    mg = RasterModelGrid((10, 10))
    mg.add_zeros("node", "topographic__elevation")
    noise = np.random.rand(mg.size("node"))
    mg.at_node["topographic__elevation"] += noise
    fr = FlowAccumulator(mg, flow_director="D8")
    fr.run_one_step()
    with pytest.raises(ValueError):
        DrainageDensity(
            mg,
            area_coefficient=1,
            slope_coefficient=1,
            area_exponent=1,
            channelization_threshold=1,
        )


def test_missing_slope_coefficient():
    mg = RasterModelGrid((10, 10))
    mg.add_zeros("node", "topographic__elevation")
    noise = np.random.rand(mg.size("node"))
    mg.at_node["topographic__elevation"] += noise
    fr = FlowAccumulator(mg, flow_director="D8")
    fr.run_one_step()
    with pytest.raises(ValueError):
        DrainageDensity(
            mg,
            area_coefficient=1,
            area_exponent=1,
            slope_exponent=1,
            channelization_threshold=1,
        )


def test_missing_area_exponent():
    mg = RasterModelGrid((10, 10))
    mg.add_zeros("node", "topographic__elevation")
    noise = np.random.rand(mg.size("node"))
    mg.at_node["topographic__elevation"] += noise
    fr = FlowAccumulator(mg, flow_director="D8")
    fr.run_one_step()
    with pytest.raises(ValueError):
        DrainageDensity(
            mg,
            area_coefficient=1,
            slope_coefficient=1,
            slope_exponent=1,
            channelization_threshold=1,
        )


def test_missing_area_coefficient():
    mg = RasterModelGrid((10, 10))
    mg.add_zeros("node", "topographic__elevation")
    noise = np.random.rand(mg.size("node"))
    mg.at_node["topographic__elevation"] += noise
    fr = FlowAccumulator(mg, flow_director="D8")
    fr.run_one_step()
    with pytest.raises(ValueError):
        DrainageDensity(
            mg,
            slope_coefficient=1,
            area_exponent=1,
            slope_exponent=1,
            channelization_threshold=1,
        )



#! /usr/env/python

"""flow_director_to_many.py provides a private class to help create
FlowDirectors.

Provides the _FlowDirectorToMany component which makes sure all model
grid fields are set up correctly.
"""
from landlab.components.flow_director.flow_director import _FlowDirector


class _FlowDirectorToMany(_FlowDirector):

    """Private class for creating components to calculate flow directions.

    This class is not meant to be used directly in modeling efforts. It
    inherits from the _FlowDirector class and builds on it to provide the
    functionality that all flow direction calculators need if they direct flow
    only to multiple nodes, as in  D infinity or MFD direction finding. It
    exists in contrast to the other intermediate flow director class
    _FlowDirectorToOne which provides equivalent functionality for flow
    direction algorithms such as D8 or steepest descent which directs flow only
    to one other node. As the primary difference between these two methods is
    the names of the fields they create and use, the primary function of this
    class is to create model grid fields.

    The primary method of this class, :func:`run_one_step` is not implemented.

    Parameters
    ----------
    grid : ModelGrid
        A grid.
    surface : field name at node or array of length node
        The surface to direct flow across.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.components.flow_director.flow_director_to_many import(
    ... _FlowDirectorToMany)
    >>> mg = RasterModelGrid((3,3), xy_spacing=(1, 1))
    >>> mg.set_closed_boundaries_at_grid_edges(True, True, True, False)
    >>> _ = mg.add_field(
    ...     'topographic__elevation',
    ...     mg.node_x + mg.node_y,
    ...     at = 'node'
    ... )
    >>> fd = _FlowDirectorToMany(mg, 'topographic__elevation')
    >>> fd.surface_values
    array([ 0.,  1.,  2.,  1.,  2.,  3.,  2.,  3.,  4.])
    >>> sorted(list(mg.at_node.keys()))
    ['flow__sink_flag', 'topographic__elevation']
    """

    _name = "FlowDirectorToMany"

    _input_var_names = ("topographic__elevation",)

    _output_var_names = ("flow__sink_flag",)

    _var_units = {"topographic__elevation": "m", "flow__sink_flag": "-"}

    _var_mapping = {"topographic__elevation": "node", "flow__sink_flag": "node"}

    _var_doc = {
        "topographic__elevation": "Land surface topographic elevation",
        "flow__sink_flag": "Boolean array, True at local lows",
    }

    def __init__(self, grid, surface):
        """Initialize the _FlowDirectorTo_One class."""
        # run init for the inherited class
        super(_FlowDirectorToMany, self).__init__(grid, surface)
        self.to_n_receivers = "many"
        # initialize new fields

    def run_one_step(self):
        """run_one_step is not implemented for this component."""
        raise NotImplementedError("run_one_step()")

    @property
    def proportions_of_flow(self):
        """Return the proportions of flow going to recievers."""
        return self._grid["node"]["flow__receiver_proportions"]


if __name__ == "__main__":  # pragma: no cover
    import doctest

    doctest.testmod()
